from typing import Tuple

import torch
import torch.nn as nn
from torch.distributions import Categorical

from velora.models.base import LiquidNCPModule, NCPModule


class SACActorDiscrete(LiquidNCPModule):
    """
    A Liquid NCP Actor Network for the SAC algorithm. Outputs a categorical
    distribution over actions.

    Usable with discrete action spaces.
    """

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(num_obs, n_neurons, num_actions, device=device)

        self.num_actions = num_actions

        self.softmax = nn.Softmax(dim=-1)

    @torch.jit.ignore
    def get_sample(self, probs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Computes a set of action samples and log probabilities using a
        Categorical distribution.

        Parameters:
            probs (torch.Tensor): Softmax probabilities for each action

        Returns:
            actions (torch.Tensor): action samples.
            log_probs (torch.Tensor): action log probabilities.
        """
        dist = Categorical(probs=probs)

        actions = dist.sample()
        log_probs = dist.log_prob(actions).unsqueeze(-1)

        return actions, log_probs

    @torch.jit.ignore
    def predict(
        self, obs: torch.Tensor, hidden: torch.Tensor | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a deterministic prediction.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            actions (torch.Tensor): sampled actions
            hidden (torch.Tensor): the new hidden state
        """
        logits, new_hidden = self.ncp(obs, hidden)
        x = self.softmax(logits)
        actions = torch.argmax(x, dim=-1)

        return actions, new_hidden

    def forward(
        self, obs: torch.Tensor, hidden: torch.Tensor | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            actions (torch.Tensor): the action predictions.
            probs (torch.Tensor): softmax probabilities for each action.
            log_prob (torch.Tensor): log probabilities of actions.
            hidden (torch.Tensor): the new hidden state.
        """
        logits, new_hidden = self.ncp(obs, hidden)
        probs = self.softmax(logits)

        actions, log_prob = self.get_sample(probs)
        return actions, probs, log_prob, new_hidden


class SACCriticDiscrete(LiquidNCPModule):
    """
    A Liquid NCP Critic Network for the SAC algorithm. Estimates Q-values given
    states and actions.

    Usable with discrete action spaces.
    """

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(num_obs, n_neurons, num_actions, device=device)

    def forward(
        self,
        obs: torch.Tensor,
        hidden: torch.Tensor | None = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            q_values (torch.Tensor): the Q-Value predictions.
            hidden (torch.Tensor): the new hidden state.
        """
        q_values, new_hidden = self.ncp(obs, hidden)
        return q_values, new_hidden


class SACCriticNCPDiscrete(NCPModule):
    """
    An NCP Critic Network for the SAC algorithm. Estimates Q-values given
    states and actions.

    Usable with discrete action spaces.
    """

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(num_obs, n_neurons, num_actions, device=device)

    def forward(self, obs: torch.Tensor) -> torch.Tensor:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations

        Returns:
            q_values (torch.Tensor): the Q-Value predictions.
        """
        q_values = self.ncp(obs)
        return q_values



from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Literal, Self, Tuple, Type, Union

from velora.callbacks import TrainCallback
from velora.training.display import training_info
from velora.training.handler import TrainHandler

try:
    from typing import override
except ImportError:  # pragma: no cover
    from typing_extensions import override  # pragma: no cover

import gymnasium as gym
import torch
import torch.nn as nn
import torch.optim as optim

if TYPE_CHECKING:
    from velora.gym.wrap import DiscreteGymNames, ContinuousGymNames  # pragma: no cover
    from velora.buffer.experience import BatchExperience  # pragma: no cover

from velora.buffer.replay import ReplayBuffer
from velora.models.base import RLModuleAgent
from velora.models.config import ModelDetails, RLAgentConfig, TorchConfig
from velora.models.nf.modules import (
    ActorModule,
    ActorModuleDiscrete,
    CriticModule,
    CriticModuleDiscrete,
    EntropyModule,
    EntropyModuleDiscrete,
)
from velora.utils.restore import load_model, save_model

StateDictKeys = Literal["modules", "optimizers"]


class NeuroFlowCT(RLModuleAgent):
    """
    ???+ abstract "Documentation"

        > [User Guide - Tutorials: NeuroFlow - Continuous](https://velora.achronus.dev/learn/tutorial/agents/nf)

    A custom Liquid RL agent that combines a variety of RL techniques.

    Designed for `continuous` action spaces.
    """

    def __init__(
        self,
        env_id: Union[str, "ContinuousGymNames"],
        actor_neurons: int,
        critic_neurons: int,
        *,
        optim: Type[optim.Optimizer] = optim.Adam,
        buffer_size: int = 1_000_000,
        actor_lr: float = 3e-4,
        critic_lr: float = 3e-4,
        alpha_lr: float = 3e-4,
        initial_alpha: float = 1.0,
        log_std: Tuple[float, float] = (-5, 2),
        tau: float = 0.005,
        gamma: float = 0.99,
        device: torch.device | None = None,
        seed: int | None = None,
    ) -> None:
        """
        Parameters:
            env_id (str): the Gymnasium environment ID to train the model on
            actor_neurons (int): number of decision nodes (inter and command nodes)
                for the actor
            critic_neurons (int): number of decision nodes (inter and command nodes)
                for the critic
            optim (Type[torch.optim.Optimizer], optional): the type of `PyTorch`
                optimizer to use
            buffer_size (int, optional): the maximum size of the `ReplayBuffer`
            actor_lr (float, optional): the actor optimizer learning rate
            critic_lr (float, optional): the critic optimizer learning rate
            alpha_lr (float, optional): the entropy parameter learning rate
            initial_alpha (float, optional): the starting entropy coefficient value
            log_std (Tuple[float, float], optional): `(low, high)` bounds for the
                log standard deviation of the action distribution. Controls the
                variance of actions
            tau (float, optional): the soft update factor used to slowly update
                the target networks
            gamma (float, optional): the reward discount factor
            device (torch.device, optional): the device to perform computations on
            seed (int, optional): random number seed for experiment
                reproducibility. When `None` generates a seed automatically
        """
        env = gym.make(env_id, render_mode="rgb_array")

        if not isinstance(env.action_space, gym.spaces.Box):
            raise EnvironmentError(
                f"Invalid '{env.action_space=}'. Must be 'gym.spaces.Box'."
            )

        super().__init__(
            env,
            actor_neurons,
            critic_neurons,
            buffer_size,
            optim,
            device,
            seed,
        )

        self.initial_alpha = initial_alpha
        self.log_std = log_std
        self.gamma = gamma
        self.tau = tau

        self.actor: ActorModule = ActorModule(
            self.state_dim,
            self.actor_neurons,
            self.action_dim,
            self.action_scale,
            self.action_bias,
            log_std_min=log_std[0],
            log_std_max=log_std[1],
            optim=optim,
            lr=actor_lr,
            device=self.device,
        )

        self.critic: CriticModule = CriticModule(
            self.state_dim,
            self.critic_neurons,
            self.action_dim,
            optim=optim,
            lr=critic_lr,
            tau=tau,
            device=self.device,
        )

        self.hidden_dim = self.actor.hidden_size

        self.entropy: EntropyModule = EntropyModule(
            self.action_dim,
            initial_alpha=initial_alpha,
            optim=optim,
            lr=alpha_lr,
            device=device,
        )

        self.loss = nn.MSELoss()
        self.buffer: ReplayBuffer = ReplayBuffer(
            buffer_size,
            self.state_dim,
            self.action_dim,
            self.actor.hidden_size,
            device=device,
        )

        self.active_params = self.actor.active_params + self.critic.active_params
        self.total_params = self.actor.total_params + self.critic.total_params

        # Init config details
        self.config = RLAgentConfig(
            agent=self.__class__.__name__,
            env=env_id,
            seed=self.seed,
            model_details=ModelDetails(
                **locals(),
                state_dim=self.state_dim,
                action_dim=self.action_dim,
                exploration_type="Entropy",
                actor=self.actor.config,
                critic=self.critic.config,
                entropy=self.entropy.config(),
            ),
            buffer=self.buffer.config(),
            torch=TorchConfig(
                device=str(self.device),
                optimizer=optim.__name__,
                loss=self.loss.__class__.__name__,
            ),
        )

        self.metadata = self.set_metadata(locals(), self.seed)

    def _update_critics(self, batch: "BatchExperience") -> torch.Tensor:
        """
        Helper method. Performs Critic network updates.

        Parameters:
            batch (BatchExperience): an object containing a batch of experience
                with `(states, actions, rewards, next_states, dones, hidden)`
                from the buffer

        Returns:
            critic_loss (torch.Tensor): total Critic network loss `(c1_loss + c2_loss)`.
        """
        with torch.no_grad():
            next_actions, next_log_probs, _ = self.actor.forward(
                batch.next_states, batch.hiddens
            )

            # Compute target Q-value
            next_q = self.critic.target_predict(batch.next_states, next_actions)
            next_q = next_q - self.entropy.alpha * next_log_probs
            target_q = batch.rewards + (1 - batch.dones) * self.gamma * next_q

        current_q1, current_q2 = self.critic.predict(batch.states, batch.actions)

        # Calculate loss
        c1_loss: torch.Tensor = self.loss(current_q1, target_q)
        c2_loss: torch.Tensor = self.loss(current_q2, target_q)
        critic_loss: torch.Tensor = c1_loss + c2_loss

        # Update critics
        self.critic.gradient_step(c1_loss, c2_loss)

        return critic_loss

    def _train_step(self, batch_size: int) -> Dict[str, torch.Tensor]:
        """
        Helper method. Performs a single training step.

        Parameters:
            batch_size (int): number of samples in a batch

        Returns:
            losses (Dict[str, torch.Tensor]): a dictionary of losses -

            - critic: the total critic loss.
            - actor: the actor loss.
            - entropy: the entropy loss.
        """
        batch = self.buffer.sample(batch_size)

        # Compute critic loss
        critic_loss = self._update_critics(batch)

        # Make predictions
        actions, log_probs, _ = self.actor.forward(batch.states, batch.hiddens)
        q1, q2 = self.critic.predict(batch.states, actions)

        # Compute actor and entropy losses
        next_q = torch.min(q1, q2)
        actor_loss = (self.entropy.alpha * log_probs - next_q).mean()
        entropy_loss = self.entropy.compute_loss(log_probs)

        # Update gradients
        self.actor.gradient_step(actor_loss)
        self.entropy.gradient_step(entropy_loss)

        # Update target networks
        self.critic.update_targets()

        return {
            "critic": critic_loss.detach(),
            "actor": actor_loss.detach(),
            "entropy": entropy_loss.detach(),
        }

    @override
    def train(
        self,
        batch_size: int,
        *,
        n_episodes: int = 10_000,
        callbacks: List["TrainCallback"] | None = None,
        log_freq: int = 10,
        display_count: int = 100,
        window_size: int = 100,
        max_steps: int = 1000,
        warmup_steps: int | None = None,
    ) -> None:
        """
        Trains the agent on a Gymnasium environment using a `ReplayBuffer`.

        Parameters:
            batch_size (int): the number of features in a single batch
            n_episodes (int, optional): the total number of episodes to train for
            callbacks (List[TrainCallback], optional): a list of training callbacks
                that are applied during the training process
            log_freq (int, optional): metric logging frequency (in episodes)
            display_count (int, optional): console training progress frequency
                (in episodes)
            window_size (int, optional): the reward moving average size
                (in episodes)
            max_steps (int, optional): the total number of steps per episode
            warmup_steps (int, optional): the number of samples to generate in the
                buffer before starting training. If `None` uses `batch_size * 2`
        """
        warmup_steps = batch_size * 2 if warmup_steps is None else warmup_steps

        # Add training details to config
        self.config = self.config.update(self._set_train_params(locals()))

        # Display console details
        self.env.reset(seed=self.seed)  # Set seed
        training_info(
            self,
            n_episodes,
            batch_size,
            window_size,
            warmup_steps,
            callbacks or [],
        )

        if warmup_steps > 0:
            self.buffer.warm(self, warmup_steps, 2 if warmup_steps < 8 else 8)

        with TrainHandler(
            self, n_episodes, max_steps, log_freq, window_size, callbacks
        ) as handler:
            for current_ep in range(1, n_episodes + 1):
                ep_reward = 0.0
                hidden = None

                state, _ = handler.env.reset()

                for current_step in range(1, max_steps + 1):
                    action, hidden = self.predict(state, hidden, train_mode=True)
                    next_state, reward, terminated, truncated, info = handler.env.step(
                        action
                    )
                    done = terminated or truncated

                    self.buffer.add(state, action, reward, next_state, done, hidden)

                    losses = self._train_step(batch_size)

                    handler.metrics.add_step(**losses)
                    handler.step(current_step)

                    state = next_state

                    if done:
                        ep_reward = info["episode"]["r"].item()

                        handler.metrics.add_episode(
                            current_ep,
                            info["episode"]["r"],
                            info["episode"]["l"],
                        )
                        break

                if current_ep % log_freq == 0 or current_ep == n_episodes:
                    handler.log(current_ep, "episode")

                if (
                    current_ep % display_count == 0
                    or current_ep == n_episodes
                    or handler.stop()
                ):
                    handler.metrics.info(current_ep)

                handler.episode(current_ep, ep_reward)

                # Terminate on early stopping
                if handler.stop():
                    break

    @override
    def predict(
        self,
        state: torch.Tensor,
        hidden: torch.Tensor | None = None,
        *,
        train_mode: bool = False,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Makes an action prediction using the Actor network.

        Parameters:
            state (torch.Tensor): the current state
            hidden (torch.Tensor, optional): the current hidden state
            train_mode (bool, optional): whether to make deterministic (when
                `False`) or stochastic (when `True`) action predictions

        Returns:
            action (torch.Tensor): the action prediction on the given state
            hidden (torch.Tensor): the Actor network's new hidden state
        """
        self.actor.eval_mode()
        with torch.no_grad():
            state = state.unsqueeze(0) if state.dim() < 2 else state

            if not train_mode:
                action, hidden = self.actor.predict(state, hidden)
            else:
                action, _, hidden = self.actor.forward(state, hidden)

        self.actor.train_mode()
        return action, hidden

    def save(
        self,
        dirpath: str | Path,
        *,
        buffer: bool = False,
        config: bool = False,
    ) -> None:
        save_model(self, dirpath, buffer=buffer, config=config)

    @classmethod
    def load(cls, dirpath: str | Path, *, buffer: bool = False) -> Self:
        return load_model(cls, dirpath, buffer=buffer)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            + ", ".join([f"{key}={val}" for key, val in self.metadata.items()])
            + ")"
        )


class NeuroFlow(RLModuleAgent):
    """
    ???+ abstract "Documentation"

        > [User Guide - Tutorials: NeuroFlow - Discrete](https://velora.achronus.dev/learn/tutorial/agents/nf2)

    A custom Liquid RL agent that combines a variety of RL techniques.

    Designed for `discrete` action spaces.
    """

    def __init__(
        self,
        env_id: Union[str, "DiscreteGymNames"],
        actor_neurons: int,
        critic_neurons: int,
        *,
        optim: Type[optim.Optimizer] = optim.Adam,
        buffer_size: int = 1_000_000,
        actor_lr: float = 3e-4,
        critic_lr: float = 3e-4,
        alpha_lr: float = 3e-4,
        initial_alpha: float = 1.0,
        tau: float = 0.005,
        gamma: float = 0.99,
        device: torch.device | None = None,
        seed: int | None = None,
    ) -> None:
        """
        Parameters:
            env_id (str): the Gymnasium environment ID to train the model on
            actor_neurons (int): number of decision nodes (inter and command nodes)
                for the actor
            critic_neurons (int): number of decision nodes (inter and command nodes)
                for the critic
            optim (Type[torch.optim.Optimizer], optional): the type of `PyTorch`
                optimizer to use
            buffer_size (int, optional): the maximum size of the `ReplayBuffer`
            actor_lr (float, optional): the actor optimizer learning rate
            critic_lr (float, optional): the critic optimizer learning rate
            alpha_lr (float, optional): the entropy parameter learning rate
            initial_alpha (float, optional): the starting entropy coefficient value
            tau (float, optional): the soft update factor used to slowly update
                the target networks
            gamma (float, optional): the reward discount factor
            device (torch.device, optional): the device to perform computations on
            seed (int, optional): random number seed for experiment
                reproducibility. When `None` generates a seed automatically
        """
        env = gym.make(env_id, render_mode="rgb_array")

        if not isinstance(env.action_space, gym.spaces.Discrete):
            raise EnvironmentError(
                f"Invalid '{env.action_space=}'. Must be 'gym.spaces.Discrete'."
            )

        super().__init__(
            env,
            actor_neurons,
            critic_neurons,
            buffer_size,
            optim,
            device,
            seed,
        )

        self.initial_alpha = initial_alpha
        self.gamma = gamma
        self.tau = tau

        self.actor: ActorModuleDiscrete = ActorModuleDiscrete(
            self.state_dim,
            self.actor_neurons,
            self.action_dim,
            optim=optim,
            lr=actor_lr,
            device=self.device,
        )

        self.critic: CriticModuleDiscrete = CriticModuleDiscrete(
            self.state_dim,
            self.critic_neurons,
            self.action_dim,
            optim=optim,
            lr=critic_lr,
            tau=tau,
            device=self.device,
        )

        self.hidden_dim = self.actor.hidden_size

        self.entropy: EntropyModuleDiscrete = EntropyModuleDiscrete(
            self.action_dim,
            initial_alpha=initial_alpha,
            optim=optim,
            lr=alpha_lr,
            device=device,
        )

        self.loss = nn.MSELoss()
        self.buffer: ReplayBuffer = ReplayBuffer(
            buffer_size,
            self.state_dim,
            1,
            self.actor.hidden_size,
            device=device,
        )

        self.active_params = self.actor.active_params + self.critic.active_params
        self.total_params = self.actor.total_params + self.critic.total_params

        # Init config details
        self.config = RLAgentConfig(
            agent=self.__class__.__name__,
            env=env_id,
            seed=self.seed,
            model_details=ModelDetails(
                **locals(),
                state_dim=self.state_dim,
                action_dim=self.action_dim,
                action_type="discrete",
                exploration_type="Entropy",
                actor=self.actor.config,
                critic=self.critic.config,
                entropy=self.entropy.config(),
            ),
            buffer=self.buffer.config(),
            torch=TorchConfig(
                device=str(self.device),
                optimizer=optim.__name__,
                loss=self.loss.__class__.__name__,
            ),
        )

        self.metadata = self.set_metadata(locals(), self.seed)

    def _update_critics(self, batch: "BatchExperience") -> torch.Tensor:
        """
        Helper method. Performs Critic network updates.

        Parameters:
            batch (BatchExperience): an object containing a batch of experience
                with `(states, actions, rewards, next_states, dones, hidden)`
                from the buffer

        Returns:
            critic_loss (torch.Tensor): total Critic network loss `(c1_loss + c2_loss)`.
        """
        with torch.no_grad():
            _, next_probs, _, _ = self.actor.forward(batch.next_states, batch.hiddens)
            next_log_probs = torch.log(next_probs + 1e-8)

            # Compute target Q-value (all actions)
            min_next_q = self.critic.target_predict(batch.next_states)  # [b, a_dim]

            next_q = next_probs * (min_next_q - self.entropy.alpha * next_log_probs)
            next_q = torch.sum(next_q, dim=-1, keepdim=True)
            target_q = batch.rewards + (1 - batch.dones) * self.gamma * next_q  # [b, 1]

        # Compute Q-value predictions for current critics (all actions)
        current_q1, current_q2 = self.critic.predict(batch.states)  # [b, a_dim]

        # Select Q-values for the actions taken in the batch - shape: (b, 1)
        current_q1 = current_q1.gather(1, batch.actions.long())
        current_q2 = current_q2.gather(1, batch.actions.long())

        # Calculate loss
        c1_loss: torch.Tensor = self.loss(current_q1, target_q)
        c2_loss: torch.Tensor = self.loss(current_q2, target_q)
        critic_loss: torch.Tensor = c1_loss + c2_loss

        # Update critics
        self.critic.gradient_step(c1_loss, c2_loss)

        return critic_loss

    def _train_step(self, batch_size: int) -> Dict[str, torch.Tensor]:
        """
        Helper method. Performs a single training step.

        Parameters:
            batch_size (int): number of samples in a batch

        Returns:
            losses (Dict[str, torch.Tensor]): a dictionary of losses -

            - critic: the total critic loss.
            - actor: the actor loss.
            - entropy: the entropy loss.
        """
        batch = self.buffer.sample(batch_size)

        # Compute critic loss
        critic_loss = self._update_critics(batch)

        # Make predictions
        _, probs, log_probs, _ = self.actor.forward(batch.states, batch.hiddens)
        q1, q2 = self.critic.predict(batch.states)

        # Compute actor and entropy losses
        next_q = torch.min(q1, q2)  # [b, a_dim]
        actor_loss = probs * (self.entropy.alpha * log_probs - next_q)
        actor_loss = torch.sum(actor_loss, dim=-1, keepdim=False).mean()

        entropy_loss = self.entropy.compute_loss(
            probs,
            torch.log(probs + 1e-8),
        )

        # Update gradients
        self.actor.gradient_step(actor_loss)
        self.entropy.gradient_step(entropy_loss)

        # Update target networks
        self.critic.update_targets()

        return {
            "critic": critic_loss.detach(),
            "actor": actor_loss.detach(),
            "entropy": entropy_loss.detach(),
        }

    @override
    def train(
        self,
        batch_size: int,
        *,
        n_episodes: int = 10_000,
        callbacks: List["TrainCallback"] | None = None,
        log_freq: int = 10,
        display_count: int = 100,
        window_size: int = 100,
        max_steps: int = 1000,
        warmup_steps: int | None = None,
    ) -> None:
        """
        Trains the agent on a Gymnasium environment using a `ReplayBuffer`.

        Parameters:
            batch_size (int): the number of features in a single batch
            n_episodes (int, optional): the total number of episodes to train for
            callbacks (List[TrainCallback], optional): a list of training callbacks
                that are applied during the training process
            log_freq (int, optional): metric logging frequency (in episodes)
            display_count (int, optional): console training progress frequency
                (in episodes)
            window_size (int, optional): the reward moving average size
                (in episodes)
            max_steps (int, optional): the total number of steps per episode
            warmup_steps (int, optional): the number of samples to generate in the
                buffer before starting training. If `None` uses `batch_size * 2`
        """
        warmup_steps = batch_size * 2 if warmup_steps is None else warmup_steps

        # Add training details to config
        self.config = self.config.update(self._set_train_params(locals()))

        # Display console details
        self.env.reset(seed=self.seed)  # Set seed
        training_info(
            self,
            n_episodes,
            batch_size,
            window_size,
            warmup_steps,
            callbacks or [],
        )

        if warmup_steps > 0:
            self.buffer.warm(self, warmup_steps, 2 if warmup_steps < 8 else 8)

        with TrainHandler(
            self, n_episodes, max_steps, log_freq, window_size, callbacks
        ) as handler:
            for current_ep in range(1, n_episodes + 1):
                ep_reward = 0.0
                hidden = None

                state, _ = handler.env.reset()

                for current_step in range(1, max_steps + 1):
                    action, hidden = self.predict(state, hidden, train_mode=True)
                    next_state, reward, terminated, truncated, info = handler.env.step(
                        action
                    )
                    done = terminated or truncated

                    self.buffer.add(state, action, reward, next_state, done, hidden)

                    losses = self._train_step(batch_size)

                    handler.metrics.add_step(**losses)
                    handler.step(current_step)

                    state = next_state

                    if done:
                        ep_reward = info["episode"]["r"].item()

                        handler.metrics.add_episode(
                            current_ep,
                            info["episode"]["r"],
                            info["episode"]["l"],
                        )
                        break

                if current_ep % log_freq == 0 or current_ep == n_episodes:
                    handler.log(current_ep, "episode")

                if (
                    current_ep % display_count == 0
                    or current_ep == n_episodes
                    or handler.stop()
                ):
                    handler.metrics.info(current_ep)

                handler.episode(current_ep, ep_reward)

                # Terminate on early stopping
                if handler.stop():
                    break

    @override
    def predict(
        self,
        state: torch.Tensor,
        hidden: torch.Tensor | None = None,
        *,
        train_mode: bool = False,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Makes an action prediction using the Actor network.

        Parameters:
            state (torch.Tensor): the current state
            hidden (torch.Tensor, optional): the current hidden state
            train_mode (bool, optional): whether to make deterministic (when
                `False`) or stochastic (when `True`) action predictions

        Returns:
            action (torch.Tensor): the action prediction on the given state
            hidden (torch.Tensor): the Actor network's new hidden state
        """
        self.actor.eval_mode()
        with torch.no_grad():
            state = state.unsqueeze(0) if state.dim() < 2 else state

            if not train_mode:
                action, hidden = self.actor.predict(state, hidden)
            else:
                action, _, _, hidden = self.actor.forward(state, hidden)

        self.actor.train_mode()
        return action, hidden

    def save(
        self,
        dirpath: str | Path,
        *,
        buffer: bool = False,
        config: bool = False,
    ) -> None:
        save_model(self, dirpath, buffer=buffer, config=config)

    @classmethod
    def load(cls, dirpath: str | Path, *, buffer: bool = False) -> Self:
        return load_model(cls, dirpath, buffer=buffer)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            + ", ".join([f"{key}={val}" for key, val in self.metadata.items()])
            + ")"
        )



import os

from sqlalchemy import Engine, ScalarResult
from sqlmodel import Session, SQLModel, create_engine, select

from velora.metrics.models import Episode


def get_db_engine() -> Engine:
    """
    Starts the metric database engine and returns it.

    Returns:
        engine (sqlalchemy.Engine): a database engine instance
    """
    from velora.metrics.models import Episode, Experiment  # pragma: no cover

    if os.getenv("VELORA_TEST_MODE", "").lower() in ("true", "1"):
        # Use in-memory database for testing
        engine = create_engine("sqlite:///:memory:")
    else:
        # Regular file-based database
        engine = create_engine("sqlite:///metrics.db")

    SQLModel.metadata.create_all(engine)
    return engine


def get_current_episode(
    session: Session,
    experiment_id: int,
    current_ep: int,
) -> ScalarResult[Episode]:
    """
    Queries a database session to retrieve the current episode for an experiment.

    Parameters:
        session (sqlmodel.Session): a metric database session
        experiment_id (int): the current experiment's unique ID
        current_ep (int): the current episode index

    Returns:
        results (ScalarResult[Episode]): a iterable set of matching episodes.
    """
    statement = select(Episode).where(
        Episode.experiment_id == experiment_id,
        Episode.episode_num == current_ep,
    )
    return session.exec(statement)



from functools import reduce
from typing import Callable, List, Literal

import gymnasium as gym
import torch

WrapperType = List[gym.Wrapper | gym.vector.VectorWrapper | Callable]


ContinuousGymNames = Literal[
    "MountainCarContinuous-v0",
    "Pendulum-v1",
    "LunarLanderContinuous-v3",
    "BipedalWalker-v3",
    "BipedalWalkerHardcore-v3",
    "CarRacing-v3",
    "Reacher-v5",
    "Pusher-v5",
    "InvertedPendulum-v5",
    "InvertedDoublePendulum-v5",
    "HalfCheetah-v5",
    "Hopper-v5",
    "Swimmer-v5",
    "Walker2d-v5",
    "Ant-v5",
    "Humanoid-v5",
    "HumanoidStandup-v5",
]

DiscreteGymNames = Literal[
    "Acrobot-v1",
    "CartPole-v1",
    "MountainCar-v0",
    "LunarLander-v3",
    "Blackjack-v1",
    "Taxi-v3",
    "CliffWalking-v0",
    "FrozenLake-v1",
    "FrozenLake8x8-v1",
]


def wrap_gym_env(
    env: gym.Env | str | DiscreteGymNames | ContinuousGymNames,
    wrappers: List[gym.Wrapper | gym.vector.VectorWrapper | Callable],
) -> gym.Env:
    """
    Creates a new [Gymnasium](https://gymnasium.farama.org/) environment with
    one or more [gymnasium.Wrappers](https://gymnasium.farama.org/api/wrappers/table/) applied.

    Parameters:
        env (gymnasium.Env | str): a name of a Gymnasium environment or the
            environment itself to wrap
        wrappers (List[gym.Wrapper | gym.vector.VectorWrapper | functools.partial]): a list of wrapper classes or partially applied wrapper functions

    Examples:
        A Gymnasium environment with normalization and reward clipping:
        ```python
        from functools import partial

        from gymnasium.wrappers import (
            NormalizeObservation,
            NormalizeReward,
            ClipReward,
        )

        env = wrap_gym_env("InvertedPendulum-v5", [
            partial(NormalizeObservation, epsilon=1e-8),
            partial(NormalizeReward, gamma=0.99, epsilon=1e-8),
            partial(ClipReward, max_reward=10.0)
        ])
        ```

    Returns:
        env (gymnasium.Env): The wrapped environment
    """
    if isinstance(env, str):
        env = gym.make(env, render_mode="rgb_array")

    def apply_wrapper(env: gym.Env, wrapper: WrapperType) -> gym.Env:
        return wrapper(env)

    return reduce(apply_wrapper, wrappers, env)


def add_core_env_wrappers(env: gym.Env, device: torch.device) -> gym.Env:
    """
    Wraps a [Gymnasium](https://gymnasium.farama.org/) environment with the following (in order) if not already applied:

    - [RecordEpisodeStatistics](https://gymnasium.farama.org/api/wrappers/misc_wrappers/#gymnasium.wrappers.RecordEpisodeStatistics) - for easily retrieving episode statistics.
    - [NumpyToTorch](https://gymnasium.farama.org/api/wrappers/misc_wrappers/#gymnasium.wrappers.NumpyToTorch) - for turning environment feedback into `PyTorch` tensors.

    Used in all pre-built algorithms.

    Parameters:
        env (gym.Env): the Gymnasium environment
        device (torch.device): the PyTorch device to perform computations on

    Returns:
        env (gym.Env): an updated Gymnasium environment
    """
    if isinstance(env, gym.Env):
        from gymnasium.wrappers import RecordEpisodeStatistics
        from gymnasium.wrappers.numpy_to_torch import NumpyToTorch
    else:
        raise ValueError(f"{type(env)}' is not supported.")

    has_stats = False
    has_torch = False

    current_env = env
    while hasattr(current_env, "env"):
        if isinstance(current_env, RecordEpisodeStatistics):
            has_stats = True
        if isinstance(current_env, NumpyToTorch):
            has_torch = True

        current_env = current_env.env

    if not has_stats:
        env = RecordEpisodeStatistics(env)

    if not has_torch:
        env = NumpyToTorch(env, device=device)

    return env



import math
from typing import Callable, Dict, Literal

import torch.nn as nn

WeightInitType = Literal[
    "xavier",
    "sonar",
    "zero",
    "trunc_normal",
    "kaiming_uniform",
    "orthogonal",
]


def get_init_fn(style: WeightInitType | None) -> Callable | None:
    """Retrieve a weight initialization function based on a given style."""
    if style is None:
        return None

    mapping: Dict[WeightInitType, Callable] = {
        "xavier": init_linear_xavier,
        "sonar": init_linear_sonar,
        "zero": init_linear_zero,
        "trunc_normal": init_linear_trunc_normal,
        "kaiming_uniform": init_linear_kaiming_uniform,
        "orthogonal": init_linear_orthogonal,
    }

    return mapping[style]


def init_linear_xavier(layer: nn.Module) -> None:
    """
    Performs Xavier weight initialization as described in the paper:
    https://proceedings.mlr.press/v9/glorot10a.html.
    """
    nn.init.xavier_normal_(layer.weight)

    if layer.bias is not None:
        nn.init.zeros_(layer.bias)


def init_linear_sonar(layer: nn.Module, sonar_std: float = 0.006) -> None:
    """
    Performs SONAR weight initialization as described in the paper:
    https://arxiv.org/abs/2308.11466.
    """
    std = sonar_std * (3 / layer.in_features) ** 0.5
    nn.init.uniform_(layer.weight, a=-std, b=std)

    if layer.bias is not None:
        nn.init.zeros_(layer.bias)


def init_linear_zero(layer: nn.Module) -> None:
    """
    Performs ZerO weight initialization as described in the paper:
    https://arxiv.org/abs/2110.12661.
    """
    nn.init.zeros_(layer.weight)

    if layer.bias is not None:
        nn.init.zeros_(layer.bias)


def init_linear_trunc_normal(layer: nn.Module) -> None:
    """
    Performs Truncated Normal weight initialization.
    """
    nn.init.trunc_normal_(layer.weight, std=1e-3)

    if layer.bias is not None:
        nn.init.zeros_(layer.bias)


def init_linear_kaiming_uniform(layer: nn.Module) -> None:
    """
    Performs Kaiming Uniform weight initialization as described in the paper:
    https://arxiv.org/abs/1502.01852.
    """
    nn.init.kaiming_uniform_(layer.weight, a=math.sqrt(5))

    if layer.bias is not None:
        fan_in, _ = nn.init._calculate_fan_in_and_fan_out(layer.weight)
        bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
        nn.init.uniform_(layer.bias, -bound, bound)


def init_linear_orthogonal(layer: nn.Module, gain: float = 1.0) -> None:
    """
    Performs Orthogonal weight initialization as described in the paper:
    https://arxiv.org/abs/1312.6120.
    """
    nn.init.orthogonal_(layer.weight, gain=gain)

    if layer.bias is not None:
        nn.init.zeros_(layer.bias)



from typing import Any, Dict, Literal, Self, Tuple

from pydantic import BaseModel


class BufferConfig(BaseModel):
    """
    A config model for buffer details.

    Attributes:
        type: the type of buffer
        capacity: the maximum capacity of the buffer
        state_dim: dimension of state observations
        action_dim: dimension of actions
        hidden_dim: dimension of hidden state
    """

    type: Literal["ReplayBuffer", "RolloutBuffer"]
    capacity: int
    state_dim: int
    action_dim: int
    hidden_dim: int


class TorchConfig(BaseModel):
    """
    A config model for PyTorch details.

    Attributes:
        device: the device used to train the model
        optimizer: the type of optimizer used
        loss: the type of optimizer used
    """

    device: str
    optimizer: str
    loss: str


class TrainConfig(BaseModel):
    """
    A config model for training parameter details.

    Attributes:
        batch_size: the size of the training batch
        n_episodes: the total number of episodes trained for. Default is `None`
        window_size: reward moving average size (in episodes)
        display_count: console training progress frequency (in episodes)
        log_freq: metric logging frequency (in episodes)
        callbacks: a dictionary of callback details. Default is `None`
        max_steps: the maximum number of steps per training episode.
            Default is `None`
        warmup_steps: number of random steps to take before starting
            training
    """

    batch_size: int
    n_episodes: int
    window_size: int
    display_count: int
    log_freq: int
    callbacks: Dict[str, Any] | None = None
    max_steps: int
    warmup_steps: int


class ModuleConfig(BaseModel):
    """
    A config model for a module's details.

    Attributes:
        active_params: active module parameters count
        total_params: total module parameter count
        architecture: a summary of the module's architecture
    """

    active_params: int
    total_params: int
    architecture: Dict[str, Any]


class SACExtraParameters(BaseModel):
    """
    A config model for extra parameters for the Soft Actor-Critic (SAC) agent.

    Attributes:
        alpha_lr: the entropy parameter learning rate
        initial_alpha: the starting entropy coefficient value
        target_entropy: the target entropy for automatic adjustment
        log_std_min: lower bound for the log standard deviation of the
            action distribution. Default is `None`
        log_std_max: upper bound for the log standard deviation of the
            action distribution. Default is `None`
    """

    alpha_lr: float
    initial_alpha: float
    target_entropy: float
    log_std_min: float | None = None
    log_std_max: float | None = None


class EntropyParameters(BaseModel):
    """
    A config model for extra parameters for NeuroFlow agents.

    Attributes:
        lr: the entropy parameter learning rate
        initial_alpha: the starting entropy coefficient value
        target: the target entropy for automatic adjustment
    """

    lr: float
    initial_alpha: float
    target: float


class CuriosityConfig(BaseModel):
    """
    A config model for the Intrinsic Curiosity Module (ICM).

    Attributes:
        icm: details about the ICM
        lr: the optimizers learning rate
        eta: importance scaling factor for intrinsic reward
        beta: weight balancing for inverse vs. forward model
    """

    icm: ModuleConfig
    lr: float
    eta: float
    beta: float


class CriticConfig(BaseModel):
    """
    A critic config model for storing a NeuroFlow agent's critic module details.

    Attributes:
        critic1: details about the first critic network
        critic2: details about the second critic network
    """

    critic1: ModuleConfig
    critic2: ModuleConfig


class ModelDetails(BaseModel):
    """
    A config model for storing an agent's network model details.

    Attributes:
        type: the type of architecture used. Default is `actor-critic`
        state_dim: number of input features
        actor_neurons: number of actor network decision nodes
        critic_neurons: number of critic network decision nodes
        action_dim: number of output features
        action_type: the type of action space. Default is `continuous`
        tau: the soft update factor for target networks
        gamma: the reward discount factor
        target_networks: whether the agent uses target networks or not.
            Default is `True`
        log_std: lower and upper bounds for the log standard deviation of the
            action distribution. Only required for `continuous` spaces.
            Default is `None`
        exploration_type: the type of agent exploration used
        actor: details about the Actor network
        critic: details about the Critic networks
        entropy: details about the entropy exploration
    """

    type: str = "actor-critic"
    state_dim: int
    actor_neurons: int
    critic_neurons: int
    action_dim: int
    tau: float
    gamma: float
    action_type: Literal["discrete", "continuous"] = "continuous"
    target_networks: bool = True
    log_std: Tuple[float, float] | None = None
    exploration_type: Literal["Entropy", "CAT-Entropy"]
    actor: ModuleConfig
    critic: CriticConfig
    entropy: EntropyParameters


class RLAgentConfig(BaseModel):
    """
    A config model for NeuroFlow agents. Stored with agent states during the
    `save()` method.

    Attributes:
        agent: the type of agent used
        env: the Gymnasium environment ID the model was trained on
        seed: random number generator value
        model_details: the agent's network model details
        buffer: the buffer details
        torch: the PyTorch details
        train_params: the agents training parameters. Default is `None`
    """

    agent: str
    env: str
    seed: int
    model_details: ModelDetails
    buffer: BufferConfig
    torch: TorchConfig
    train_params: TrainConfig | None = None

    def update(self, train_params: TrainConfig) -> Self:
        """
        Updates the training details of the model.

        Parameters:
            train_params (TrainConfig): a config containing training parameters

        Returns:
            self (Self): a new config model with the updated values.
        """
        return RLAgentConfig(
            train_params=train_params,
            **self.model_dump(exclude={"train_params"}),
        )



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Dict, Tuple, Type

import torch
import torch.nn as nn
import torch.optim as optim

from velora.models.config import CriticConfig, EntropyParameters
from velora.models.sac.continuous import SACActor, SACCriticNCP
from velora.models.sac.discrete import SACActorDiscrete, SACCriticNCPDiscrete
from velora.utils.torch import soft_update


class BaseModule(ABC):
    """
    A base module for all agent modules.
    """

    @abstractmethod
    def state_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        Creates a state dictionary for the module.

        Returns:
            state_dict (Dict[str, Dict[str, Any]]): the state dicts for the module, including networks and optimizers.
        """
        pass  # pragma: no cover

    @abstractmethod
    def load_state_dict(self, state_dict: Dict[str, Dict[str, Any]]) -> None:
        """
        Load the modules state dict from a previously saved state.

        Parameters:
            state_dict (Dict[str, Dict[str, Any]]): a previously saved state dict
        """
        pass  # pragma: no cover


class ActorModule(BaseModule):
    """
    An Actor module for NeuroFlow. Uses a Liquid NCP SAC Actor with a
    Gaussian policy.

    Usable with continuous action spaces.
    """

    def __init__(
        self,
        state_dim: int,
        n_neurons: int,
        action_dim: int,
        action_scale: torch.Tensor,
        action_bias: torch.Tensor,
        *,
        log_std_min: float = -5,
        log_std_max: float = 2,
        optim: Type[optim.Optimizer] = optim.Adam,
        lr: float = 3e-4,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            state_dim (int): dimension of the state space
            n_neurons (int): number of decision/hidden neurons
            action_dim (int): dimension of the action space
            action_scale (torch.Tensor): scale factor to map normalized actions to
                environment's action range
            action_bias (torch.Tensor): bias/offset to center normalized actions to
                environment's action range
            log_std_min (float, optional): minimum log standard deviation
            log_std_max (float, optional): maximum log standard deviation
            optim (Type[optim.Optimizer], optional): a `PyTorch` optimizer class
            lr (float, optional): optimizer learning rate
            device (torch.device, optional): the device to perform computations on
        """
        self.state_dim = state_dim
        self.n_neurons = n_neurons
        self.action_dim = action_dim
        self.action_scale = action_scale
        self.action_bias = action_bias
        self.log_std = (log_std_min, log_std_max)
        self.lr = lr
        self.device = device

        self.network = SACActor(
            state_dim,
            n_neurons,
            action_dim,
            action_scale,
            action_bias,
            log_std_min=log_std_min,
            log_std_max=log_std_max,
            device=device,
        ).to(device)

        self.hidden_size = self.network.ncp.hidden_size

        self.optim = optim(self.network.parameters(), lr=lr)

        self.config = self.network.config()

        self.active_params = self.config.active_params
        self.total_params = self.config.total_params

        self.network: SACActor = torch.jit.script(self.network)

    def gradient_step(self, loss: torch.Tensor) -> None:
        """
        Performs a gradient update step.

        Parameters:
            loss (torch.Tensor): network loss
        """
        self.optim.zero_grad()
        loss.backward()
        self.optim.step()

    def predict(
        self,
        obs: torch.Tensor,
        hidden: torch.Tensor | None = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Makes a deterministic prediction using the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the current hidden state

        Returns:
            actions (torch.Tensor): the action predictions.
            hidden (torch.Tensor): the new hidden state.
        """
        action, hidden = self.network.predict(obs, hidden)
        return action, hidden

    def forward(
        self, obs: torch.Tensor, hidden: torch.Tensor | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the current hidden state

        Returns:
            actions (torch.Tensor): the action predictions.
            log_prob (torch.Tensor): log probabilities of actions.
            hidden (torch.Tensor): the new hidden state.
        """
        action, log_prob, hidden = self.network(obs, hidden)
        return action, log_prob, hidden

    def eval_mode(self) -> None:
        """Sets the network to evaluation mode."""
        self.network.eval()

    def train_mode(self) -> None:
        """Sets the network to training mode."""
        self.network.train()

    def state_dict(self) -> Dict[str, Dict[str, Any]]:
        return {
            "actor": self.network.state_dict(),
            "actor_optim": self.optim.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Dict[str, Any]]) -> None:
        self.network.load_state_dict(state_dict["actor"])
        self.optim.load_state_dict(state_dict["actor_optim"])

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"state_dim={self.state_dim}, "
            f"n_neurons={self.n_neurons}, "
            f"action_dim={self.action_dim}, "
            f"action_scale={self.action_scale}, "
            f"action_bias={self.action_bias}, "
            f"optim={type(self.optim).__name__}, "
            f"log_std={self.log_std}, "
            f"lr={self.lr}, "
            f"device={self.device})"
        )


class ActorModuleDiscrete(BaseModule):
    """
    An Actor module for NeuroFlow. Uses a Liquid NCP SAC Actor with a
    Categorical policy.

    Usable with discrete action spaces.
    """

    def __init__(
        self,
        state_dim: int,
        n_neurons: int,
        action_dim: int,
        *,
        optim: Type[optim.Optimizer] = optim.Adam,
        lr: float = 3e-4,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            state_dim (int): dimension of the state space
            n_neurons (int): number of decision/hidden neurons
            action_dim (int): dimension of the action space
            optim (Type[optim.Optimizer], optional): a `PyTorch` optimizer class
            lr (float, optional): optimizer learning rate
            device (torch.device, optional): the device to perform computations on
        """
        self.state_dim = state_dim
        self.n_neurons = n_neurons
        self.action_dim = action_dim
        self.lr = lr
        self.device = device

        self.network = SACActorDiscrete(
            state_dim,
            n_neurons,
            action_dim,
            device=device,
        ).to(device)

        self.hidden_size = self.network.ncp.hidden_size

        self.optim = optim(self.network.parameters(), lr=lr)

        self.config = self.network.config()

        self.active_params = self.config.active_params
        self.total_params = self.config.total_params

        self.network: SACActorDiscrete = torch.jit.script(self.network)

    def gradient_step(self, loss: torch.Tensor) -> None:
        """
        Performs a gradient update step.

        Parameters:
            loss (torch.Tensor): network loss
        """
        self.optim.zero_grad()
        loss.backward()
        self.optim.step()

    def predict(
        self,
        obs: torch.Tensor,
        hidden: torch.Tensor | None = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Makes a deterministic prediction using the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the current hidden state

        Returns:
            actions (torch.Tensor): the action predictions.
            hidden (torch.Tensor): the new hidden state.
        """
        action, hidden = self.network.predict(obs, hidden)
        return action, hidden

    def forward(
        self, obs: torch.Tensor, hidden: torch.Tensor | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the current hidden state

        Returns:
            actions (torch.Tensor): the action predictions.
            probs (torch.Tensor): softmax probabilities for each action.
            log_prob (torch.Tensor): log probabilities of actions.
            hidden (torch.Tensor): the new hidden state.
        """
        actions, probs, log_prob, hidden = self.network(obs, hidden)
        return actions, probs, log_prob, hidden

    def eval_mode(self) -> None:
        """Sets the network to evaluation mode."""
        self.network.eval()

    def train_mode(self) -> None:
        """Sets the network to training mode."""
        self.network.train()

    def state_dict(self) -> Dict[str, Dict[str, Any]]:
        return {
            "actor": self.network.state_dict(),
            "actor_optim": self.optim.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Dict[str, Any]]) -> None:
        self.network.load_state_dict(state_dict["actor"])
        self.optim.load_state_dict(state_dict["actor_optim"])

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"state_dim={self.state_dim}, "
            f"n_neurons={self.n_neurons}, "
            f"action_dim={self.action_dim}, "
            f"optim={type(self.optim).__name__}, "
            f"lr={self.lr}, "
            f"device={self.device})"
        )


class CriticModule(BaseModule):
    """
    A Critic module for NeuroFlow. Uses a pair of NCP SAC Critic's with separate
    target networks to estimate Q-values.

    Usable with continuous action spaces.
    """

    def __init__(
        self,
        state_dim: int,
        n_neurons: int,
        action_dim: int,
        *,
        optim: Type[optim.Optimizer] = optim.Adam,
        lr: float = 3e-4,
        tau: float = 0.005,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            state_dim (int): dimension of the state space
            n_neurons (int): number of decision/hidden neurons
            action_dim (int): dimension of the action space
            optim (Type[optim.Optimizer], optional): a `PyTorch` optimizer class
            lr (float, optional): optimizer learning rates
            tau (float, optional): soft target network update factor
            device (torch.device, optional): the device to perform computations on
        """
        self.state_dim = state_dim
        self.n_neurons = n_neurons
        self.action_dim = action_dim
        self.lr = lr
        self.tau = tau
        self.device = device

        self.network1 = SACCriticNCP(
            state_dim,
            n_neurons,
            action_dim,
            device=device,
        ).to(device)

        self.network2 = SACCriticNCP(
            state_dim,
            n_neurons,
            action_dim,
            device=device,
        ).to(device)

        self.target1 = deepcopy(self.network1)
        self.target2 = deepcopy(self.network2)

        self.optim1 = optim(self.network1.parameters(), lr=lr)
        self.optim2 = optim(self.network2.parameters(), lr=lr)

        self.config = CriticConfig(
            critic1=self.network1.config(),
            critic2=self.network2.config(),
        )

        self.active_params = (
            self.config.critic1.active_params + self.config.critic2.active_params
        )
        self.total_params = (
            self.config.critic1.total_params + self.config.critic2.total_params
        )

        self.network1: SACCriticNCP = torch.jit.script(self.network1)
        self.network2: SACCriticNCP = torch.jit.script(self.network2)

        self.target1: SACCriticNCP = torch.jit.script(self.target1)
        self.target2: SACCriticNCP = torch.jit.script(self.target2)

    def update_targets(self) -> None:
        """
        Performs a soft update on the target networks.
        """
        soft_update(self.network1, self.target1, tau=self.tau)
        soft_update(self.network2, self.target2, tau=self.tau)

    def gradient_step(self, c1_loss: torch.Tensor, c2_loss: torch.Tensor) -> None:
        """
        Performs a gradient update step.

        Parameters:
            c1_loss (torch.Tensor): critic loss for first network
            c2_loss (torch.Tensor): critic loss for second network
        """
        self.optim1.zero_grad()
        c1_loss.backward()
        self.optim1.step()

        self.optim2.zero_grad()
        c2_loss.backward()
        self.optim2.step()

    def predict(
        self, obs: torch.Tensor, actions: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Makes a prediction using the critic networks.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            actions (torch.Tensor): the batch of actions

        Returns:
            q_values1 (torch.Tensor): the Q-Value predictions from the first network.
            q_values2 (torch.Tensor): the Q-Value predictions from the second network.
        """
        q_values1 = self.network1(obs, actions)
        q_values2 = self.network2(obs, actions)

        return q_values1, q_values2

    def target_predict(self, obs: torch.Tensor, actions: torch.Tensor) -> torch.Tensor:
        """
        Makes a prediction using the target networks.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            actions (torch.Tensor): the batch of actions

        Returns:
            next_q (torch.Tensor): the smallest Q-Value predictions between the target networks.
        """
        q_values1 = self.target1(obs, actions)
        q_values2 = self.target2(obs, actions)

        return torch.min(q_values1, q_values2)

    def state_dict(self) -> Dict[str, Dict[str, Any]]:
        return {
            "critic": self.network1.state_dict(),
            "critic2": self.network2.state_dict(),
            "critic_target": self.target1.state_dict(),
            "critic2_target": self.target2.state_dict(),
            "critic_optim": self.optim1.state_dict(),
            "critic2_optim": self.optim2.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Dict[str, Any]]) -> None:
        self.network1.load_state_dict(state_dict["critic"])
        self.network2.load_state_dict(state_dict["critic2"])

        self.target1.load_state_dict(state_dict["critic_target"])
        self.target2.load_state_dict(state_dict["critic2_target"])

        self.optim1.load_state_dict(state_dict["critic_optim"])
        self.optim2.load_state_dict(state_dict["critic2_optim"])

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"state_dim={self.state_dim}, "
            f"n_neurons={self.n_neurons}, "
            f"action_dim={self.action_dim}, "
            f"optim={type(self.optim1).__name__}, "
            f"lr={self.lr}, "
            f"tau={self.tau}, "
            f"device={self.device})"
        )


class CriticModuleDiscrete(BaseModule):
    """
    A Critic module for NeuroFlow. Uses a pair of NCP SAC Critic's with separate
    target networks to estimate Q-values.

    Usable with discrete action spaces.
    """

    def __init__(
        self,
        state_dim: int,
        n_neurons: int,
        action_dim: int,
        *,
        optim: Type[optim.Optimizer] = optim.Adam,
        lr: float = 3e-4,
        tau: float = 0.005,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            state_dim (int): dimension of the state space
            n_neurons (int): number of decision/hidden neurons
            action_dim (int): dimension of the action space
            optim (Type[optim.Optimizer], optional): a `PyTorch` optimizer class
            lr (float, optional): optimizer learning rates
            tau (float, optional): soft target network update factor
            device (torch.device, optional): the device to perform computations on
        """
        self.state_dim = state_dim
        self.n_neurons = n_neurons
        self.action_dim = action_dim
        self.lr = lr
        self.tau = tau
        self.device = device

        self.network1 = SACCriticNCPDiscrete(
            state_dim,
            n_neurons,
            action_dim,
            device=device,
        ).to(device)

        self.network2 = SACCriticNCPDiscrete(
            state_dim,
            n_neurons,
            action_dim,
            device=device,
        ).to(device)

        self.target1 = deepcopy(self.network1)
        self.target2 = deepcopy(self.network2)

        self.optim1 = optim(self.network1.parameters(), lr=lr)
        self.optim2 = optim(self.network2.parameters(), lr=lr)

        self.config = CriticConfig(
            critic1=self.network1.config(),
            critic2=self.network2.config(),
        )

        self.active_params = (
            self.config.critic1.active_params + self.config.critic2.active_params
        )
        self.total_params = (
            self.config.critic1.total_params + self.config.critic2.total_params
        )

        self.network1: SACCriticNCPDiscrete = torch.jit.script(self.network1)
        self.network2: SACCriticNCPDiscrete = torch.jit.script(self.network2)

        self.target1: SACCriticNCPDiscrete = torch.jit.script(self.target1)
        self.target2: SACCriticNCPDiscrete = torch.jit.script(self.target2)

    def update_targets(self) -> None:
        """
        Performs a soft update on the target networks.
        """
        soft_update(self.network1, self.target1, tau=self.tau)
        soft_update(self.network2, self.target2, tau=self.tau)

    def gradient_step(self, c1_loss: torch.Tensor, c2_loss: torch.Tensor) -> None:
        """
        Performs a gradient update step.

        Parameters:
            c1_loss (torch.Tensor): critic loss for first network
            c2_loss (torch.Tensor): critic loss for second network
        """
        self.optim1.zero_grad()
        c1_loss.backward()
        self.optim1.step()

        self.optim2.zero_grad()
        c2_loss.backward()
        self.optim2.step()

    def predict(self, obs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Makes a prediction using the critic networks.

        Parameters:
            obs (torch.Tensor): the batch of state observations

        Returns:
            q_values1 (torch.Tensor): the Q-Value predictions from the first network.
            q_values2 (torch.Tensor): the Q-Value predictions from the second network.
        """
        q_values1 = self.network1(obs)
        q_values2 = self.network2(obs)

        return q_values1, q_values2

    def target_predict(self, obs: torch.Tensor) -> torch.Tensor:
        """
        Makes a prediction using the target networks.

        Parameters:
            obs (torch.Tensor): the batch of state observations

        Returns:
            next_q (torch.Tensor): the smallest Q-Value predictions between the target networks.
        """
        q_values1 = self.target1(obs)
        q_values2 = self.target2(obs)

        return torch.min(q_values1, q_values2)

    def state_dict(self) -> Dict[str, Dict[str, Any]]:
        return {
            "critic": self.network1.state_dict(),
            "critic2": self.network2.state_dict(),
            "critic_target": self.target1.state_dict(),
            "critic2_target": self.target2.state_dict(),
            "critic_optim": self.optim1.state_dict(),
            "critic2_optim": self.optim2.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Dict[str, Any]]) -> None:
        self.network1.load_state_dict(state_dict["critic"])
        self.network2.load_state_dict(state_dict["critic2"])

        self.target1.load_state_dict(state_dict["critic_target"])
        self.target2.load_state_dict(state_dict["critic2_target"])

        self.optim1.load_state_dict(state_dict["critic_optim"])
        self.optim2.load_state_dict(state_dict["critic2_optim"])

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"state_dim={self.state_dim}, "
            f"n_neurons={self.n_neurons}, "
            f"action_dim={self.action_dim}, "
            f"optim={type(self.optim1).__name__}, "
            f"lr={self.lr}, "
            f"tau={self.tau}, "
            f"device={self.device})"
        )


class EntropyModule(BaseModule):
    """
    An Entropy module for NeuroFlow. Uses automatic entropy tuning from SAC
    based on the paper: [Soft Actor-Critic Algorithms and Applications](https://arxiv.org/abs/1812.05905).

    Usable with continuous action spaces.
    """

    def __init__(
        self,
        action_dim: int,
        *,
        initial_alpha: float = 1.0,
        optim: Type[optim.Optimizer] = optim.Adam,
        lr: float = 3e-4,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            action_dim (int): dimension of the action space
            initial_alpha (float, optional): the starting entropy coefficient value
            optim (Type[optim.Optimizer], optional): a `PyTorch` optimizer class
            lr (float, optional): optimizer learning rates
            device (torch.device, optional): the device to perform computations on
        """
        self.action_dim = action_dim
        self.initial_alpha = initial_alpha
        self.lr = lr
        self.device = device

        self.target = -action_dim
        self.log_alpha = nn.Parameter(torch.tensor(initial_alpha, device=device).log())

        self.optim = optim([self.log_alpha], lr=lr)

    @property
    def alpha(self) -> torch.Tensor:
        """
        Get the current entropy coefficient (alpha).

        Returns:
            alpha (torch.Tensor): the entropy coefficient.
        """
        return self.log_alpha.exp()

    def compute_loss(self, log_probs: torch.Tensor) -> torch.Tensor:
        """
        Computes the entropy coefficient loss.

        Parameters:
            log_probs (torch.Tensor): log probabilities for actions

        Returns:
            loss (torch.Tensor): the entropy loss value.
        """
        loss = torch.tensor(0.0, device=log_probs.device)

        entropy = (log_probs + self.target).detach()
        loss = -(self.log_alpha * entropy).mean()

        return loss

    def gradient_step(self, loss: torch.Tensor) -> None:
        """
        Performs a gradient update step.

        Parameters:
            loss (torch.Tensor): loss to backpropagate
        """
        self.optim.zero_grad()
        loss.backward()
        self.optim.step()

    def config(self) -> EntropyParameters:
        """
        Creates a module config.

        Returns:
            config (EntropyParameters): a parameter config model.
        """
        return EntropyParameters(
            lr=self.lr,
            initial_alpha=self.initial_alpha,
            target=self.target,
        )

    def state_dict(self) -> Dict[str, Dict[str, Any]]:
        return {
            "entropy_optim": self.optim.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Dict[str, Any]]) -> None:
        self.optim.load_state_dict(state_dict["entropy_optim"])

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"action_dim={self.action_dim}, "
            f"initial_alpha={self.initial_alpha}, "
            f"optim={type(self.optim).__name__}, "
            f"lr={self.lr}, "
            f"device={self.device})"
        )


class EntropyModuleDiscrete(BaseModule):
    """
    An Entropy module for NeuroFlow. Uses automatic entropy tuning from SAC
    based on the paper: [Soft Actor-Critic for Discrete Action Settings](https://arxiv.org/abs/1910.07207).

    Usable with discrete action spaces.
    """

    def __init__(
        self,
        action_dim: int,
        *,
        initial_alpha: float = 1.0,
        optim: Type[optim.Optimizer] = optim.Adam,
        lr: float = 3e-4,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            action_dim (int): dimension of the action space
            initial_alpha (float, optional): the starting entropy coefficient value
            optim (Type[optim.Optimizer], optional): a `PyTorch` optimizer class
            lr (float, optional): optimizer learning rates
            device (torch.device, optional): the device to perform computations on
        """
        self.action_dim = action_dim
        self.initial_alpha = initial_alpha
        self.lr = lr
        self.device = device

        self.target = 0.98 * torch.tensor(1 / action_dim, device=device).log()
        self.log_alpha = nn.Parameter(torch.tensor(initial_alpha, device=device).log())

        self.optim = optim([self.log_alpha], lr=lr)

    @property
    def alpha(self) -> torch.Tensor:
        """
        Get the current entropy coefficient (alpha).

        Returns:
            alpha (torch.Tensor): the entropy coefficient.
        """
        return self.log_alpha.exp()

    def compute_loss(
        self, probs: torch.Tensor, log_probs: torch.Tensor
    ) -> torch.Tensor:
        """
        Computes the entropy coefficient loss.

        Parameters:
            probs (torch.Tensor): probabilities for actions
            log_probs (torch.Tensor): log probabilities for actions

        Returns:
            loss (torch.Tensor): the entropy loss value.
        """
        with torch.no_grad():
            entropy_mean = -torch.sum(probs * log_probs, dim=1).mean()

        loss = self.log_alpha * (entropy_mean - self.target)
        return loss

    def gradient_step(self, loss: torch.Tensor) -> None:
        """
        Performs a gradient update step.

        Parameters:
            loss (torch.Tensor): loss to backpropagate
        """
        self.optim.zero_grad()
        loss.backward()
        self.optim.step()

    def config(self) -> EntropyParameters:
        """
        Creates a module config.

        Returns:
            config (EntropyParameters): a parameter config model.
        """
        return EntropyParameters(
            lr=self.lr,
            initial_alpha=self.initial_alpha,
            target=self.target,
        )

    def state_dict(self) -> Dict[str, Dict[str, Any]]:
        return {
            "entropy_optim": self.optim.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Dict[str, Any]]) -> None:
        self.optim.load_state_dict(state_dict["entropy_optim"])

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"action_dim={self.action_dim}, "
            f"initial_alpha={self.initial_alpha}, "
            f"optim={type(self.optim).__name__}, "
            f"lr={self.lr}, "
            f"device={self.device})"
        )



import re
from dataclasses import dataclass
from typing import Dict, List, Literal

import gymnasium as gym


@dataclass
class EnvResult:
    """
    Container for environment search results.

    Parameters:
        name (str): the name of the environment
        type (str): the environment type, either `discrete` or `continuous`
    """

    name: str
    type: str


class SearchHelper:
    """
    A helper class containing methods for creating the `EnvSearch` cache.
    """

    @staticmethod
    def get_env_type(env_name: str) -> Literal["discrete", "continuous"] | None:
        """
        Determines if an environment has a discrete or continuous action space.

        Parameters:
            env_name (str): name of the environment

        Returns:
            env_type (str | None): one of three values -

            - `discrete` for Discrete action space.
            - `continuous` for Box action space.
            - `None`, otherwise.
        """
        try:
            env = gym.make(env_name)

            if isinstance(env.action_space, gym.spaces.Discrete):
                return "discrete"
            elif isinstance(env.action_space, gym.spaces.Box):
                return "continuous"
            return None

        except Exception as e:
            raise e
        finally:
            if "env" in locals():
                env.close()

    @staticmethod
    def get_latest_env_names() -> List[str]:
        """
        Builds a list of the latest [Gymnasium](https://gymnasium.farama.org/) environment names.

        Returns:
            names (List[str]): a list of names for all latest env versions.
        """
        env_dict: Dict[str, int] = {}

        for env_name in gym.envs.registry.keys():
            # Skip envs with paths (JAX-based) or old Gym versions
            if "/" in env_name or env_name.startswith("GymV"):
                continue

            match = re.match(r"(.+)-v(\d+)$", env_name)

            if match:
                base_name, version = match.groups()
                version = int(version)

            # Keep only the latest version
            if base_name not in env_dict or version > env_dict[base_name]:
                env_dict[base_name] = version

        return [f"{name}-v{version}" for name, version in env_dict.items()]


class EnvSearch:
    """
    A utility class for searching for [Gymnasium](https://gymnasium.farama.org/)
    environments.
    """

    _result_cache: List[EnvResult] = []
    _helper = SearchHelper()

    @classmethod
    def _build_name_cache(cls) -> None:
        """
        Builds the environment name cache storing the latest
        [Gymnasium](https://gymnasium.farama.org/) environment names.
        """
        if cls._result_cache:
            return

        names = cls._helper.get_latest_env_names()
        print(f"Caching {len(names)} environments...", end="")
        results = []

        for env in names:
            env_type = cls._helper.get_env_type(env)
            results.append(EnvResult(name=env, type=env_type))

        cls._result_cache = results
        print("Complete.")

    @classmethod
    def find(cls, query: str) -> List[EnvResult]:
        """
        Find a [Gymnasium](https://gymnasium.farama.org/) environment that contains `query`.

        Parameters:
            query (str): partial or complete name of an environment
                (e.g., `Lunar` or `Pendulum`)

        Returns:
            result (List[EnvResult]): a list of environment results matching the query.
        """
        cls._build_name_cache()

        matches = [env for env in cls._result_cache if query in env.name]
        print(f"{len(matches)} environments found.")
        return matches

    @classmethod
    def discrete(cls) -> List[EnvResult]:
        """
        Get all available discrete [Gymnasium](https://gymnasium.farama.org/) environments.

        Returns:
            names (List[EnvResult]): a list of available discrete environments.
        """
        cls._build_name_cache()

        matches = [env for env in cls._result_cache if env.type == "discrete"]
        print(f"{len(matches)} environments found.")
        return matches

    @classmethod
    def continuous(cls) -> List[EnvResult]:
        """
        Get all available continuous [Gymnasium](https://gymnasium.farama.org/) environments.

        Returns:
            names (List[EnvResult]): a list of available continuous environments.
        """
        cls._build_name_cache()

        matches = [env for env in cls._result_cache if env.type == "continuous"]
        print(f"{len(matches)} environments found.")
        return matches



import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Tuple

import gymnasium as gym
import torch

from velora.gym.wrap import add_core_env_wrappers

if TYPE_CHECKING:
    from velora.models.base import RLModuleAgent  # pragma: no cover


def record_episode(agent: "RLModuleAgent", dirpath: str | Path | None = None) -> None:
    """
    Manually makes a video recording of an agent in an episode.

    Parameters:
        agent (RLModuleAgent): an agent to use
        dirpath (str, optional): a path to save the video. When `None` uses the
            current working directory
    """
    dirpath = Path(dirpath) if dirpath else Path().cwd()

    def trigger(t: int) -> bool:
        return True

    env = gym.make(agent.env.spec.id, render_mode="rgb_array")

    # Ignore folder warning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        env = gym.wrappers.RecordVideo(
            env,
            dirpath,
            episode_trigger=trigger,
            name_prefix=f"{env.spec.name}_final",
        )

    env = add_core_env_wrappers(env, device=agent.device)

    for _ in range(1):
        hidden = None
        done = False

        state, _ = env.reset()

        while not done:
            action, hidden = agent.predict(state, hidden, train_mode=False)

            next_state, _, terminated, truncated, _ = env.step(action)
            done = terminated or truncated

            state = next_state

    env.close()


def evaluate_agent(agent: "RLModuleAgent") -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Evaluates the performance of an agent on an environment using a single episode.

    Parameters:
        agent (RLModuleAgent): the agent to evaluate

    Returns:
        ep_return (torch.Tensor): episodic return.
        ep_length (torch.Tensor): episode length.
    """
    env = agent.eval_env

    state, _ = env.reset()
    hidden = None

    ep_return = 0
    ep_length = 0

    while True:
        action, hidden = agent.predict(state, hidden, train_mode=False)
        next_state, reward, terminated, truncated, info = env.step(action.flatten())

        done = terminated or truncated
        state = next_state

        if done:
            ep_return = info["episode"]["r"]
            ep_length = info["episode"]["l"]
            break

    return ep_return, ep_length



import json
from pathlib import Path
from pydoc import locate
from typing import TYPE_CHECKING, Any, Dict, Literal, Type, get_args

if TYPE_CHECKING:
    from velora.models.base import RLModuleAgent  # pragma: no cover

import torch
from safetensors.torch import load_file, save_file

TensorDictKeys = Literal[
    "actor",
    "critic",
    "critic2",
    "critic_target",
    "critic2_target",
]
MetadataKeys = Literal[
    "model",
    "actor_optim",
    "critic_optim",
    "critic2_optim",
    "entropy_optim",
    "buffer",
]
OptimDictKeys = Literal["actor_optim", "critic_optim", "critic2_optim", "entropy_optim"]
ModuleNames = Literal["actor", "critic", "entropy"]


def optim_to_tensor(name: str, state_dict: Dict[str, Any]) -> Dict[str, torch.Tensor]:
    """
    Extracts an optimizers `state` from it's `state_dict()` and converts it into
    a `tensor_dict` for storage.

    Parameters:
        name (str): name of the optimizer
        state_dict (Dict[str, Any]): PyTorch optimizer state dictionary

    Returns:
        tensor_dict (Dict[str, torch.Tensor]): the converted state as a tensor dictionary
    """
    tensor_dict: Dict[str, torch.Tensor] = {}

    if "state" in state_dict:
        for param_id, param_state in state_dict["state"].items():
            for k, v in param_state.items():
                tensor_key = f"{name}.{param_id}.{k}"
                tensor_dict[tensor_key] = v.cpu()

    return tensor_dict


def optim_from_tensor(tensor_dict: Dict[str, torch.Tensor]) -> Dict[OptimDictKeys, Any]:
    """
    Converts an optimizer's `tensor_dict` back into a `state_dict()` without the
    metadata.

    Parameters:
        tensor_dict (Dict[str, torch.Tensor]): PyTorch optimizers tensor dictionary

    Returns:
        state_dict (Dict[str, Any]): the converted state as a normal dictionary
    """
    state_dict: Dict[OptimDictKeys, Any] = {}

    for key, tensor in tensor_dict.items():
        optim_name, param_id, param_key = key.split(".")
        param_id = int(param_id)

        if optim_name not in state_dict:
            state_dict[optim_name] = {}

        if param_id not in state_dict[optim_name]:
            state_dict[optim_name][param_id] = {}

        state_dict[optim_name][param_id][param_key] = tensor

    return state_dict


def model_from_tensor(
    tensor_dict: Dict[str, torch.Tensor],
) -> Dict[TensorDictKeys, Any]:
    """
    Converts a model's `tensor_dict` back into a `state_dict()`.

    Parameters:
        tensor_dict (Dict[str, torch.Tensor]): PyTorch model tensor dictionary

    Returns:
        state_dict (Dict[str, Any]): the converted state as a normal dictionary
    """
    state_dict: Dict[TensorDictKeys, Any] = {}

    for key, tensor in tensor_dict.items():
        model_name, param_name = key.split(".", 1)

        if model_name not in state_dict:
            state_dict[model_name] = {}

        state_dict[model_name][param_name] = tensor

    return state_dict


def save_model(
    agent: "RLModuleAgent",
    dirpath: str | Path,
    *,
    buffer: bool = False,
    config: bool = False,
    force: bool = False,
) -> None:
    """
    Saves the current model state into `safetensors` and `json` files.

    !!! warning

        `model_config.json` is stored in the `dirpath.parent`.

    Includes:
    - `model_config.json` - contains the core details of the agent (optional)
    - `metadata.json` - contains the model, optimizer and buffer (optional) metadata
    - `model_state.safetensors` - contains the model weights and biases
    - `optim_state.safetensors` - contains the optimizer states (actor and critic)
    - `buffer_state.safetensors` - contains the buffer state (only if `buffer=True`)

    Parameters:
        agent (RLModuleAgent): the agent's state to save
        dirpath (str | Path): the location to store the model state. Should only
            consist of `folder` names. E.g., `<folder>/<folder>`
        buffer (bool, optional): a flag for storing the buffer state
        config (bool, optional): a flag for storing the model's config
        force (bool, optional): enables file overwriting, ignoring existing state
            files. Useful for continuing model training
    """
    save_path = Path(dirpath)
    save_path.mkdir(parents=True, exist_ok=True)

    config_path = Path(save_path.parent, "model_config").with_suffix(".json")
    metadata_path = Path(save_path, "metadata").with_suffix(".json")
    model_state_path = Path(save_path, "model_state").with_suffix(".safetensors")
    optim_state_path = Path(save_path, "optim_state").with_suffix(".safetensors")
    buffer_state_path = Path(save_path, "buffer_state").with_suffix(".safetensors")

    if not force and model_state_path.exists():
        raise FileExistsError(
            f"A model state already exists in the '{save_path}' directory! Either change the 'dirpath', delete the folders contents, or use 'force=True' to allow overwriting."
        )

    # Handle module state dicts
    agent_state = agent.state_dict()
    tensor_dict: Dict[str, torch.Tensor] = {}

    for model_name, state_dict in agent_state["modules"].items():
        for param_name, tensor in state_dict.items():
            tensor_dict[f"{model_name}.{param_name}"] = tensor.contiguous()

    # Handle optimizers
    optim_dict: Dict[OptimDictKeys, Dict[str, Any]] = {}
    param_dict: Dict[OptimDictKeys, Dict[str, Any]] = {}

    for optim_name, state_dict in agent_state["optimizers"].items():
        optim_dict |= optim_to_tensor(optim_name, state_dict)
        param_dict[optim_name] = state_dict["param_groups"]

    metadata: Dict[MetadataKeys, Any] = {
        "model": agent.metadata,
        **param_dict,
    }

    # Save tensors (weights and biases only)
    save_file(tensor_dict, model_state_path)
    save_file(optim_dict, optim_state_path)

    # Add buffer (if applicable)
    if buffer:
        metadata["buffer"] = agent.buffer.metadata()
        save_file(agent.buffer.state_dict(), buffer_state_path)

    # Write to files
    if config and not config_path.exists():
        with config_path.open("w") as f:
            f.write(agent.config.model_dump_json(indent=2, exclude_none=True))

    if not metadata_path.exists():
        with metadata_path.open("w") as f:
            f.write(json.dumps(metadata, indent=2))


def load_model(
    agent: Type["RLModuleAgent"], dirpath: str | Path, *, buffer: bool = False
) -> Any:
    """
    Creates a new agent instance by loading a saved one from the `dirpath`.
    Also, loads the original training buffer if `buffer=True`.

    These files must exist in the `dirpath`:
    - `metadata.json` - contains the model, optimizer and buffer (optional) metadata
    - `model_state.safetensors` - contains the model weights and biases
    - `optim_state.safetensors` - contains the optimizer states (actor and critic)
    - `buffer_state.safetensors` - contains the buffer state (only if `buffer=True`)

    Parameters:
        agent (Type[RLModuleAgent]): the type of agent to load
        dirpath (str | Path): the location to store the model state. Should only
            consist of `folder` names. E.g., `<folder>/<folder>`
        buffer (bool, optional): a flag for storing the buffer state

    Returns:
        agent (RLModuleAgent): a new agent instance with the saved state
    """
    load_path = Path(dirpath)

    metadata_path = Path(load_path, "metadata").with_suffix(".json")
    model_state_path = Path(load_path, "model_state").with_suffix(".safetensors")
    optim_state_path = Path(load_path, "optim_state").with_suffix(".safetensors")
    buffer_state_path = Path(load_path, "buffer_state").with_suffix(".safetensors")

    if not model_state_path.exists():
        raise FileNotFoundError(f"Model state '{model_state_path}' does not exist!")

    if not optim_state_path.exists():
        raise FileNotFoundError(f"Optimizer state '{optim_state_path}' does not exist!")

    if not metadata_path.exists():
        raise FileNotFoundError(f"Metadata '{metadata_path}' does not exist!")

    if buffer and not buffer_state_path.exists():
        raise FileNotFoundError(
            f"Buffer state '{buffer_state_path}' does not exist! Try with 'buffer=False'."
        )

    # Load metadata first
    with metadata_path.open("r") as f:
        metadata: Dict[MetadataKeys, Any] = json.load(f)

    device: str = metadata["model"]["device"]
    metadata["model"]["device"] = torch.device(device)
    metadata["model"]["optim"] = locate(metadata["model"]["optim"])

    # Create new model instance
    model = agent(**metadata["model"])

    # Load module and optimizer parameters from safetensors
    tensor_dict: Dict[str, torch.Tensor] = load_file(model_state_path, device)
    agent_state = model_from_tensor(tensor_dict)

    tensor_dict: Dict[str, torch.Tensor] = load_file(optim_state_path, device)
    optim_state = optim_from_tensor(tensor_dict)

    metadata.pop("model")
    for key in metadata.keys():
        state = optim_state[key] if key in optim_state else {}

        agent_state[key] = {
            "state": state,
            "param_groups": metadata[key],
        }

    # Restore module states
    for name in get_args(ModuleNames):
        module = getattr(model, name)
        module.load_state_dict(agent_state)

    # Load buffer
    if buffer:
        model.buffer = model.buffer.load(buffer_state_path, metadata["buffer"])

    print(f"Loaded model:\n  {model}\n  buffer_restored={buffer}")
    return model



try:
    from typing import override
except ImportError:  # pragma: no cover
    from typing_extensions import override  # pragma: no cover

from typing import TYPE_CHECKING

import gymnasium as gym
import torch

if TYPE_CHECKING:
    from velora.models.base import RLModuleAgent  # pragma: no cover

from velora.buffer.base import BufferBase
from velora.buffer.experience import BatchExperience
from velora.models.config import BufferConfig


class ReplayBuffer(BufferBase):
    """
    A Buffer for storing agent experiences. Used for Off-Policy agents.

    First introduced in Deep RL in the Deep Q-Network paper:
    [Player Atari with Deep Reinforcement Learning](https://arxiv.org/abs/1312.5602).
    """

    def __init__(
        self,
        capacity: int,
        state_dim: int,
        action_dim: int,
        hidden_dim: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            capacity (int): the total capacity of the buffer
            state_dim (int): dimension of state observations
            action_dim (int): dimension of actions
            hidden_dim (int): dimension of hidden state
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(capacity, state_dim, action_dim, hidden_dim, device=device)

    def config(self) -> BufferConfig:
        """
        Creates a buffer config model.

        Returns:
            config (BufferConfig): a config model with buffer details.
        """
        return BufferConfig(
            type="ReplayBuffer",
            capacity=self.capacity,
            state_dim=self.state_dim,
            action_dim=self.action_dim,
            hidden_dim=self.hidden_dim,
        )

    @override
    def sample(self, batch_size: int) -> BatchExperience:
        """
        Samples a random batch of experiences from the buffer.

        Parameters:
            batch_size (int): the number of items to sample

        Returns:
            batch (BatchExperience): an object of samples with the attributes (`states`, `actions`, `rewards`, `next_states`, `dones`, `hidden`).

                All items have the same shape `(batch_size, features)`.
        """
        if len(self) < batch_size:
            raise ValueError(
                f"Buffer does not contain enough experiences. Available: {len(self)}, Requested: {batch_size}"
            )

        indices = torch.randint(0, self.size, (batch_size,), device=self.device)

        return BatchExperience(
            states=self.states[indices],
            actions=self.actions[indices],
            rewards=self.rewards[indices],
            next_states=self.next_states[indices],
            dones=self.dones[indices],
            hiddens=self.hiddens[indices],
        )

    def warm(self, agent: "RLModuleAgent", n_samples: int, num_envs: int = 8) -> None:
        """
        Warms the buffer to fill it to a number of samples by generating them
        from an agent using a `vectorized` copy of the environment.

        Parameters:
            agent (Any): the agent to generate samples with
            n_samples (int): the maximum number of samples to generate
            num_envs (int, optional): number of vectorized environments. Cannot
                be smaller than `2`
        """
        if num_envs < 2:
            raise ValueError(f"'{num_envs=}' cannot be smaller than 2.")

        envs = gym.make_vec(
            agent.env.spec.id,
            num_envs=num_envs,
            vectorization_mode="sync",
        )
        envs: gym.vector.SyncVectorEnv = gym.wrappers.vector.NumpyToTorch(
            envs, agent.device
        )

        hidden = None
        states, _ = envs.reset()

        while not len(self) >= n_samples:
            actions, hidden = agent.predict(states, hidden, train_mode=True)
            next_states, rewards, terminated, truncated, _ = envs.step(actions)
            dones = terminated | truncated

            self.add_multi(states, actions, rewards, next_states, dones, hidden)

            states = next_states

        envs.close()



from velora.models.lnn import LiquidNCPNetwork, NCPNetwork
from velora.models.nf import NeuroFlowCT, NeuroFlow

__all__ = [
    "LiquidNCPNetwork",
    "NCPNetwork",
    "NeuroFlowCT",
    "NeuroFlow",
]



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from velora.models.config import RLAgentConfig  # pragma: no cover

import torch
from sqlmodel import Session

from velora.metrics.models import Episode, Experiment
from velora.utils.format import number_to_short


class StepStorage:
    """
    A storage container for step metrics.

    Useful for calculating the episodic average values to store in `MetricStorage`.

    Attributes:
        critic_losses (torch.Tensor): a tensor of agent Critic loss values
        actor_losses (torch.Tensor): a tensor of agent Actor loss values
        entropy_losses (torch.Tensor): a tensor of agent Entropy loss values
    """

    def __init__(self, capacity: int, *, device: torch.device | None = None) -> None:
        """
        Parameters:
            capacity (int): storage capacity for each tensor
            device (torch.device, optional): the device to perform computations on
        """
        self.capacity = capacity
        self.device = device

        # Position indicators
        self.position = 0
        self.size = 0

        self.critic_losses = torch.zeros((capacity), device=device)
        self.actor_losses = torch.zeros((capacity), device=device)
        self.entropy_losses = torch.zeros((capacity), device=device)

    def critic_avg(self, ep_length: int) -> torch.Tensor:
        """
        Computes the critic loss average. Useful for computing episodic averages.

        Parameters:
            ep_length (int): size of the episode

        Returns:
            avg (torch.Tensor): critic loss step average
        """
        return self.critic_losses[:ep_length].mean()

    def actor_avg(self, ep_length: int) -> torch.Tensor:
        """
        Computes the actor loss average. Useful for computing episodic averages.

        Parameters:
            ep_length (int): size of the episode

        Returns:
            avg (torch.Tensor): actor loss step average
        """
        return self.actor_losses[:ep_length].mean()

    def entropy_avg(self, ep_length: int) -> torch.Tensor:
        """
        Computes the entropy loss average. Useful for computing episodic averages.

        Parameters:
            ep_length (int): size of the episode

        Returns:
            avg (torch.Tensor): entropy loss step average
        """
        return self.entropy_losses[:ep_length].mean()

    def add(
        self,
        critic: torch.Tensor,
        actor: torch.Tensor,
        entropy: torch.Tensor,
    ) -> None:
        """
        Adds one of each metric into storage.

        Parameters:
            critic (torch.Tensor): critic loss
            actor (torch.Tensor): actor loss
            entropy (torch.Tensor): entropy loss
        """
        self.critic_losses[self.position] = critic
        self.actor_losses[self.position] = actor
        self.entropy_losses[self.position] = entropy

        # Update position
        self.position = (self.position + 1) % self.capacity
        self.size = min(self.size + 1, self.capacity)

    def empty(self) -> None:
        """Empty storage."""
        self.critic_losses.zero_()
        self.actor_losses.zero_()
        self.entropy_losses.zero_()

        self.position = 0
        self.size = 0


class MovingMetric:
    """
    Tracks a metric with a moving window for statistics.

    Attributes:
        window (torch.Tensor): a list of values for the statistics
        window_size (int): the window size of the moving statistics
    """

    def __init__(self, window_size: int, *, device: torch.device | None = None) -> None:
        """
        Parameters:
            window_size (int): the size of the moving window
            device (torch.device, optional): the device to perform computations on
        """
        self.window_size = window_size
        self.device = device

        # Position indicators
        self.position = 0
        self.size = 0

        # Pre-allocated storage
        self.window = torch.zeros((window_size), device=device)

    @property
    def latest(self) -> torch.Tensor:
        """Gets the latest value."""
        latest_pos = (self.position - 1) % self.window_size
        return self.window[latest_pos]

    def add(self, value: torch.Tensor) -> None:
        """
        Adds a value and updates the window.

        Parameters:
            value (torch.Tensor): value to add
        """
        self.window[self.position] = value.to(self.device)

        # Update position - deque style
        self.position = (self.position + 1) % self.window_size
        self.size = min(self.size + 1, self.window_size)

    def mean(self) -> torch.Tensor:
        """
        Calculates the mean of values or the current window.

        Returns:
            avg (torch.Tensor): the calculated mean.
        """
        return self.window.mean()

    def std(self) -> torch.Tensor:
        """
        Calculates the standard deviation of values or the current window.

        Returns:
            std (torch.Tensor): the calculated standard deviation.
        """
        return (
            self.window.std()
            if self.window.size(dim=0) > 1
            else torch.tensor(0.0, device=self.device)
        )

    def max(self) -> torch.Tensor:
        """
        Calculates the maximum value of a set of values or the current window.

        Returns:
            max (torch.Tensor): the maximum value.
        """
        return self.window.max()

    def __len__(self) -> int:
        """Returns the number of items in the values array."""
        return self.size


class TrainMetricsBase:
    """
    A base class for training metrics.
    """

    def __init__(
        self,
        session: Session,
        window_size: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            session (sqlmodel.Session): current metric database session
            window_size (int): moving average window size
            device (torch.device, optional): the device to perform computations on
        """
        self.session = session
        self.window_size = window_size
        self.device = device

        self._ep_rewards = MovingMetric(window_size, device=device)
        self._ep_lengths = MovingMetric(window_size, device=device)

        self.experiment_id: int | None = None

        self._critic_loss: torch.Tensor = torch.zeros(1, device=self.device)
        self._actor_loss: torch.Tensor = torch.zeros(1, device=self.device)
        self._entropy_loss: torch.Tensor = torch.zeros(1, device=self.device)

        self.step_total: torch.Tensor = torch.zeros(
            1,
            dtype=torch.int32,
            device=self.device,
        )

    def start_experiment(self, config: "RLAgentConfig") -> None:
        """
        Confirms the start of a metric experiment by adding it to the database and
        storing its unique ID locally.

        Parameters:
            config (RLAgentConfig): an RLAgent config model
        """
        exp = Experiment(
            agent=config.agent,
            env=config.env,
            config=config.model_dump_json(),
        )

        self.session.add(exp)
        self.session.commit()

        self.session.refresh(exp)
        self.experiment_id = exp.id

    def reward_moving_avg(self) -> float:
        """
        Calculates the average moving reward.

        Returns:
            avg (float): the average moving reward.
        """
        return self._ep_rewards.mean().item()

    def reward_moving_std(self) -> float:
        """
        Calculates the average reward moving standard deviation.

        Returns:
            avg (float): the average moving standard deviation.
        """
        return self._ep_rewards.std().item()

    def reward_moving_max(self) -> float:
        """
        Calculates the highest reward for the window.

        Returns:
            max (float): the highest reward in the window.
        """
        return self._ep_rewards.max().item()

    def _exp_created_check(self) -> None:
        """
        Helper method. Performs error handling for checking if an experiment
        has been created first.

        Used in `add_step` and `add_episode`.
        """
        if not self.experiment_id:
            raise RuntimeError(
                "An experiment must be created first!\nCreate one with the '<TrainHandler_instance>.metrics.add_experiment()' method."
            )


class TrainMetrics(TrainMetricsBase):
    """
    A utility class for working with and storing episodic training metrics for
    monitoring an agents training performance.
    """

    def __init__(
        self,
        session: Session,
        window_size: int,
        n_episodes: int,
        max_steps: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            session (sqlmodel.Session): current metric database session
            window_size (int): moving average window size
            n_episodes (int): total number of training episodes
            max_steps (int): maximum number of steps per episode
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(session, window_size, device=device)

        self.n_episodes = n_episodes
        self.max_steps = max_steps

        self._current_losses = StepStorage(max_steps, device=device)

    def add_step(
        self,
        critic: torch.Tensor,
        actor: torch.Tensor,
        entropy: torch.Tensor,
    ) -> None:
        """
        Add timestep metrics to local storage.

        Parameters:
            critic (torch.Tensor): critic step loss
            actor (torch.Tensor): actor step loss
            entropy (torch.Tensor): entropy step loss
        """
        self._exp_created_check()

        self._current_losses.add(critic, actor, entropy)

    def add_episode(
        self,
        ep_idx: int,
        reward: torch.Tensor,
        ep_length: torch.Tensor,
    ) -> None:
        """
        Add episode metrics to the metric database and reset step accumulators.

        Parameters:
            ep_idx (int): the current episode index
            reward (torch.Tensor): episode reward
            ep_length (torch.Tensor): number of steps after episode done
        """
        self._exp_created_check()

        self._ep_rewards.add(reward.to(self.device))
        self._ep_lengths.add(ep_length.to(self.device))

        self._actor_loss = self._current_losses.actor_avg(ep_length.item())
        self._critic_loss = self._current_losses.critic_avg(ep_length.item())
        self._entropy_loss = self._current_losses.entropy_avg(ep_length.item())
        self.step_total += ep_length

        moving_avg = self.reward_moving_avg()
        moving_std = self.reward_moving_std()

        ep = Episode(
            experiment_id=self.experiment_id,
            episode_num=ep_idx,
            reward=reward.item(),
            length=ep_length.item(),
            reward_moving_avg=moving_avg,
            reward_moving_std=moving_std,
            actor_loss=self._actor_loss.item(),
            critic_loss=self._critic_loss.item(),
            entropy_loss=self._entropy_loss.item(),
        )
        self.session.add(ep)
        self.session.commit()

        # Reset step storage
        self._current_losses.empty()

    def info(self, current_ep: int) -> None:
        """
        Outputs basic information to the console.

        Parameters:
            current_ep (int): the current episode index
        """
        ep = number_to_short(current_ep)
        max_eps = number_to_short(self.n_episodes)

        ep_length = number_to_short(int(self._ep_lengths.latest))
        step_total = number_to_short(self.step_total.item())

        max_length = number_to_short(int(self._ep_lengths.max().item()))
        max_steps = number_to_short(self.max_steps)

        print(
            f"Episode: {ep}/{max_eps}, "
            f"Steps: {ep_length}/{step_total}, "
            f"Max Length: {max_length}/{max_steps}, "
            f"Reward Avg: {self.reward_moving_avg():.2f}, "
            f"Reward Max: {self.reward_moving_max():.2f}, "
            f"Actor Loss: {self._actor_loss.item():.2f}, "
            f"Critic Loss: {self._critic_loss.item():.2f}, "
            f"Entropy Loss: {self._entropy_loss.item():.2f}"
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from velora.models.weight import WeightInitType, get_init_fn


class SparseLinear(nn.Module):
    """A `torch.nn.Linear` layer with sparsely weighted connections."""

    bias: torch.Tensor
    mask: torch.Tensor

    def __init__(
        self,
        in_features: int,
        out_features: int,
        mask: torch.Tensor,
        *,
        init_type: str | WeightInitType = "kaiming_uniform",
        bias: bool = True,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            in_features (int): number of input features
            out_features (int): number of output features
            mask (torch.Tensor): sparsity mask tensor of shape
                `(out_features, in_features)`
            init_type (str, optional): the type of weight initialization
            bias (bool, optional): a flag to enable additive bias
            device (torch.device, optional): device to perform computations on
        """
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.device = device

        self.register_buffer("mask", mask.to(device).detach())

        weight = torch.empty((out_features, in_features), device=device)
        self.weight = nn.Parameter(weight)

        if bias:
            self.bias = nn.Parameter(torch.empty(out_features, device=device))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters(init_type)

        with torch.no_grad():
            self.weight.data.mul_(self.mask)

    def reset_parameters(self, style: str | WeightInitType) -> None:
        """
        Initializes weights and biases using an initialization method.
        """
        weight_fn = get_init_fn(style)
        weight_fn(self)

    def update_mask(self, mask: torch.Tensor) -> None:
        """
        Updates the sparsity mask with a new one.

        Parameters:
            mask (torch.Tensor): new mask
        """
        self.mask = mask.to(self.device).detach()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Perform a forward pass through the layer.

        Parameters:
            x (torch.Tensor): input tensor with shape `(..., in_features)`

        Returns:
            y_pred (torch.Tensor): layer prediction with sparsity applied with shape `(..., out_features)`.
        """
        return F.linear(x, self.weight * self.mask, self.bias)

    def extra_repr(self) -> str:
        """String representation of layer parameters."""
        return f"in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"



from typing import Tuple

import torch
from torch.distributions import Normal

from velora.models.base import LiquidNCPModule, NCPModule


class SACActor(LiquidNCPModule):
    """
    A Liquid NCP Actor Network for the SAC algorithm. Outputs a Gaussian
    distribution over actions.

    Usable with continuous action spaces.
    """

    action_scale: torch.Tensor
    action_bias: torch.Tensor

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        action_scale: torch.Tensor,
        action_bias: torch.Tensor,
        *,
        log_std_min: float = -5,
        log_std_max: float = 2,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            action_scale (torch.Tensor): scale factor to map normalized actions to
                environment's action range
            action_bias (torch.Tensor): bias/offset to center normalized actions to
                environment's action range
            log_std_min (float, optional): lower bound for the log standard
                deviation of the action distribution. Controls the minimum
                variance of actions
            log_std_max (float, optional): upper bound for the log standard
                deviation of the action distribution. Controls the maximum
                variance of actions
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(num_obs, n_neurons, num_actions * 2, device=device)

        self.log_std_min = log_std_min
        self.log_std_max = log_std_max

        self.register_buffer("action_scale", action_scale)
        self.register_buffer("action_bias", action_bias)

    @torch.jit.ignore
    def get_sample(
        self, mean: torch.Tensor, std: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Computes a set of action samples and log probabilities using the
        reparameterization trick from a Gaussian distribution.

        Parameters:
            mean (torch.Tensor): network prediction means.
            std (torch.Tensor): network standard deviation predictions.

        Returns:
            actions (torch.Tensor): action samples.
            log_probs (torch.Tensor): log probabilities.
        """
        dist = Normal(mean, std)
        x_t = dist.rsample()  # Reparameterization trick
        log_probs = dist.log_prob(x_t)

        return x_t, log_probs

    @torch.jit.ignore
    def predict(
        self, obs: torch.Tensor, hidden: torch.Tensor | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a deterministic prediction.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            actions (torch.Tensor): sampled actions
            hidden (torch.Tensor): the new hidden state
        """
        x, new_hidden = self.ncp(obs, hidden)

        mean, _ = torch.chunk(x, 2, dim=-1)

        # Bound actions between [-1, 1]
        actions_normalized = torch.tanh(mean)

        # Scale back to env action space
        actions = actions_normalized * self.action_scale + self.action_bias

        return actions, new_hidden

    def forward(
        self, obs: torch.Tensor, hidden: torch.Tensor | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            actions (torch.Tensor): the action predictions.
            log_prob (torch.Tensor): log probabilities of actions.
            hidden (torch.Tensor): the new hidden state.
        """
        x, new_hidden = self.ncp(obs, hidden)

        # Split output into mean and log_std
        mean, log_std = torch.chunk(x, 2, dim=-1)

        # Bound between [-20, 2]
        log_std = torch.clamp(log_std, self.log_std_min, self.log_std_max)
        std = log_std.exp()

        # Sample from normal distribution
        x_t, dist_log_probs = self.get_sample(mean, std)
        actions_normalized = torch.tanh(x_t)  # Bounded: [-1, 1]

        # Scale back to environment's action space
        actions = actions_normalized * self.action_scale + self.action_bias

        # Calculate log probability, accounting for tanh
        log_prob = dist_log_probs - torch.log(
            self.action_scale * (1 - actions_normalized.pow(2)) + 1e-6
        )
        log_prob = log_prob.sum(dim=-1, keepdim=True)

        return actions, log_prob, new_hidden


class SACCritic(LiquidNCPModule):
    """
    A Liquid NCP Critic Network for the SAC algorithm. Estimates Q-values given
    states and actions.

    Usable with continuous action spaces.
    """

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(num_obs + num_actions, n_neurons, 1, device=device)

    def forward(
        self,
        obs: torch.Tensor,
        actions: torch.Tensor,
        hidden: torch.Tensor | None = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            actions (torch.Tensor): the batch of actions
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            q_values (torch.Tensor): the Q-Value predictions.
            hidden (torch.Tensor): the new hidden state.
        """
        inputs = torch.cat([obs, actions], dim=-1)
        q_values, new_hidden = self.ncp(inputs, hidden)
        return q_values, new_hidden


class SACCriticNCP(NCPModule):
    """
    An NCP Critic Network for the SAC algorithm. Estimates Q-values given
    states and actions.

    Usable with continuous action spaces.
    """

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__(num_obs + num_actions, n_neurons, 1, device=device)

    def forward(self, obs: torch.Tensor, actions: torch.Tensor) -> torch.Tensor:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            actions (torch.Tensor): the batch of actions

        Returns:
            q_values (torch.Tensor): the Q-Value predictions.
        """
        inputs = torch.cat([obs, actions], dim=-1)
        q_values = self.ncp(inputs)
        return q_values



from velora.gym.search import EnvResult, EnvSearch, SearchHelper
from velora.gym.wrap import add_core_env_wrappers, wrap_gym_env

__all__ = [
    "add_core_env_wrappers",
    "wrap_gym_env",
    "EnvSearch",
    "EnvResult",
    "SearchHelper",
]



import json
import time
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, List, Literal, Self, Type

from sqlmodel import Session

from velora.utils.format import number_to_short

if TYPE_CHECKING:
    from velora.callbacks import TrainCallback  # pragma: no cover

from velora.gym.wrap import add_core_env_wrappers
from velora.metrics.db import get_db_engine
from velora.models.base import RLModuleAgent
from velora.state import TrainState
from velora.time import ElapsedTime
from velora.training.metrics import TrainMetrics, TrainMetricsBase
from velora.utils.capture import record_episode


class TrainHandlerBase:
    """
    A base class for train handlers.
    """

    def __init__(
        self,
        agent: RLModuleAgent,
        window_size: int,
        callbacks: List["TrainCallback"] | None,
    ) -> None:
        """
        Parameters:
            agent (RLAgent): the agent being trained
            window_size (int): episode window size rate
            callbacks (List[TrainCallback] | None): a list of training callbacks.
                If `None` sets to an empty list
        """
        self.agent = agent
        self.env = self.agent.env
        self.window_size = window_size
        self.callbacks = callbacks or []
        self.device = self.agent.device

        self.state: TrainState | None = None

        self.start_time = 0.0
        self.train_time: ElapsedTime | None = None

        self.engine = get_db_engine()
        self.session: Session | None = None
        self._metrics: TrainMetricsBase | None = None

    def __enter__(self) -> Self:
        """
        Setup the training context, initializing the environment.

        Returns:
            self (Self): the initialized context.
        """
        self.start_time = time.time()

        self.start()
        self.env = add_core_env_wrappers(self.env, self.device)

        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """
        Clean up resources and finalize training.

        Parameters:
            exc_type (Type[BaseException], optional): the exception class, if an
                exception was raised inside the `with` block. `None` otherwise
            exc_val (BaseException, optional): the exception instance, if an
                exception is raised. `None` otherwise
            exc_tb (TracebackType, optional): the traceback object, if an exception
                occurred. `None` otherwise
        """
        self.record_last_episode()

        self.complete()

        self.session.close()

        self.train_time = ElapsedTime.elapsed(self.start_time)

        early_stop_str = "Early stopping target reached!\n" if self.stop() else ""
        print(
            "---------------------------------\n"
            f"{early_stop_str}"
            "Training completed in: "
            f"{number_to_short(self.state.current_ep)} episodes, "
            f"{number_to_short(self._metrics.step_total.item())} steps, "
            f"and {self.train_time}."
        )

    def _run_callbacks(self) -> None:
        """Helper method. Runs the callbacks and updates the training state."""
        for cb in self.callbacks:
            self.state = cb(self.state)

    def complete(self) -> None:
        """Performs `complete` callback event."""
        self.state.status = "complete"
        self._run_callbacks()

    def start(self) -> None:
        """
        Performs `start` callback event.
        """
        self._run_callbacks()

    def stop(self) -> bool:
        """
        Checks if training should be stopped, such as early stopping.

        Returns:
            stop (bool): `True` if training should be stopped, `False` otherwise.
        """
        return self.state.stop_training

    def episode(self, current_ep: int, ep_reward: float) -> None:
        """
        Performs `episode` callback event.

        Parameters:
            current_ep (int): the current training episode index
            ep_reward (float): the episodes reward (return)
        """
        self.state.update(
            status="episode",
            current_ep=current_ep,
            ep_reward=ep_reward,
        )
        self._run_callbacks()

    def record_last_episode(self) -> None:
        """
        If recording videos enabled, captures a recording of the last episode.

        Only works when `TrainCallback.RecordVideos` is applied.

        Filename format: `<env_name>_final-episode-0.mp4`.
        """
        if self.state.record_state is not None:
            dirname = self.state.record_state.dirpath.parent.name
            cp_path = Path("checkpoints", dirname, "videos")
            print()
            record_episode(self.agent, cp_path)

    def save_completed(self) -> None:
        """
        Saves the completed training information to a JSON file in the
        local directory.
        """
        path = Path(self.state.checkpoint_dir, "completed").with_suffix(".json")

        # data to save
        data = {
            "episodes": self.state.current_ep,
            "steps": self._metrics.step_total.item(),
            "time": {
                "hours": self.train_time.hrs,
                "minutes": self.train_time.mins,
                "seconds": round(self.train_time.secs, 4),
            },
            "early_stopped": self.state.stop_training,
            "stats": {
                "reward": {
                    "average": round(self._metrics.reward_moving_avg(), 4),
                    "max": round(self._metrics.reward_moving_max(), 4),
                },
                "loss": {
                    "actor": round(self._metrics._actor_loss.item(), 4),
                    "critic": round(self._metrics._critic_loss.item(), 4),
                    "entropy": round(self._metrics._entropy_loss.item(), 4),
                },
            },
        }

        with path.open("w") as f:
            f.write(json.dumps(data, indent=2))


class TrainHandler(TrainHandlerBase):
    """
    A context manager for handling an agents training state. Compatible with single
    environments.
    """

    def __init__(
        self,
        agent: RLModuleAgent,
        n_episodes: int,
        max_steps: int,
        log_freq: int,
        window_size: int,
        callbacks: List["TrainCallback"] | None,
    ) -> None:
        """
        Parameters:
            agent (RLModuleAgent): the agent being trained
            n_episodes (int): the total number of training episodes
            max_steps (int): maximum number of steps in an episode
            log_freq (int): metric logging frequency (in episodes)
            window_size (int): episode window size rate
            callbacks (List[TrainCallback] | None): a list of training callbacks.
                If `None` sets to an empty list
        """
        super().__init__(agent, window_size, callbacks)

        self.log_freq = log_freq
        self.n_episodes = n_episodes
        self.max_steps = max_steps

    @property
    def metrics(self) -> TrainMetrics:
        """
        Training metric class instance.

        Returns:
            metrics (TrainMetrics): current training metric state.
        """
        return self._metrics

    def __enter__(self) -> Self:
        """
        Setup the training context, initializing the environment.

        Returns:
            self (Self): the initialized context.
        """
        self.session = Session(self.engine)
        self._metrics = TrainMetrics(
            self.session,
            self.window_size,
            self.n_episodes,
            self.max_steps,
            device=self.device,
        )
        self._metrics.start_experiment(self.agent.config)

        self.state = TrainState(
            agent=self.agent,
            env=self.env,
            session=self.session,
            total_episodes=self.n_episodes,
            experiment_id=self._metrics.experiment_id,
        )

        return super().__enter__()

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ):
        super().__exit__(exc_type, exc_val, exc_tb)

        if self.state.saving_enabled:
            self.save_completed()

    def start(self) -> None:
        super().start()

        # Update environment with callback wrappers
        self.env = self.state.env

    def step(self, current_step: int) -> None:
        """
        Performs `step` callback event.

        Parameters:
            current_step (int): the current training timestep index
        """
        self.state.update(status="step", current_step=current_step)
        self._run_callbacks()

    def log(self, idx: int, log_type: Literal["episode", "step"]) -> None:
        """
        Performs `logging` callback event.

        Parameters:
            idx (int): the current training step or episode index
            log_type (str): the type of logging method
        """
        if log_type == "episode":
            self.state.update(status="logging", current_ep=idx, logging_type=log_type)
        else:
            self.state.update(status="logging", current_step=idx, logging_type=log_type)

        self._run_callbacks()



from velora.models.sac.continuous import SACActor, SACCritic, SACCriticNCP
from velora.models.sac.discrete import (
    SACActorDiscrete,
    SACCriticDiscrete,
    SACCriticNCPDiscrete,
)

__all__ = [
    "SACActor",
    "SACCritic",
    "SACCriticNCP",
    "SACActorDiscrete",
    "SACCriticDiscrete",
    "SACCriticNCPDiscrete",
]



def number_to_short(value: int) -> str:
    """
    Converts a number into a human-readable format like `1M` or `1.25K`.

    Parameters:
        value (int): The number to convert

    Returns:
        str: The shortened version as a string
    """
    is_negative = value < 0
    abs_value = abs(value)

    suffixes = [(1_000_000_000, "B"), (1_000_000, "M"), (1_000, "K")]

    for threshold, suffix in suffixes:
        if abs_value >= threshold:
            short_value = round(abs_value / threshold, 2)
            result = f"{short_value:g}{suffix}"
            return f"-{result}" if is_negative else result

    return str(value)



from velora.buffer.base import BufferBase
from velora.buffer.experience import BatchExperience
from velora.buffer.replay import ReplayBuffer

__all__ = [
    "BufferBase",
    "BatchExperience",
    "ReplayBuffer",
]



from typing import Tuple

import torch
import torch.nn as nn

from velora.models.lnn.sparse import SparseLinear
from velora.models.weight import WeightInitType


class NCPLiquidCell(nn.Module):
    """
    A Liquid Time-Constant (LTC) cell using a Closed-form (CfC) approach.

    The LTC cell follows the closed-form continuous-depth
    (CFC; Equation 10) solution from the paper:
    [Closed-form Continuous-time Neural Models](https://arxiv.org/abs/2106.13898).

    Equation:
    $$
    x(t) =
        \\sigma(-f(x, I, θ_f), t) \\; g(x, I, θ_g)
        + \\left[ 1 - \\sigma(-[\\;f(x, I, θ_f)\\;]\\;t) \\right] \\; h(x, I, θ_h)
    $$
    """

    sparsity_mask: torch.Tensor

    def __init__(
        self,
        in_features: int,
        n_hidden: int,
        mask: torch.Tensor,
        *,
        init_type: str | WeightInitType = "kaiming_uniform",
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            in_features (int): number of input nodes.
            n_hidden (int): number of hidden nodes.
            mask (torch.Tensor): a matrix of sparse connections
                usually containing a combination of `[-1, 1, 0]`.
            init_type (str, optional): the type of weight initialization
            device (torch.device, optional): the device to load tensors on.
        """

        super().__init__()

        self.in_features = in_features
        self.n_hidden = n_hidden
        self.head_size = n_hidden + in_features
        self.init_type = init_type
        self.device = device

        # Absolute to maintain masking (-1 -> 1)
        self.register_buffer("sparsity_mask", self._prep_mask(mask.to(device)))

        self.tanh = nn.Tanh()  # Bounded: [-1, 1]
        self.sigmoid = nn.Sigmoid()  # Bounded: [0, 1]

        self.g_head = self._make_layer()
        self.h_head = self._make_layer()

        # LTC heads (f)
        self.f_head_to_g = self._make_layer()
        self.f_head_to_h = self._make_layer()

        # Hidden state projection
        self.proj = self._make_layer()

    def _make_layer(self) -> SparseLinear:
        """
        Helper method. Creates a new `SparseLinear` layer with the following values:

        - `in_features` - `self.n_hidden + self.in_features`.
        - `out_features` - `self.n_hidden`.
        - `mask` - `self.sparsity_mask`.
        - `device` - `self.device`.

        Returns:
            layer (SparseLinear): a `SparseLinear` layer.
        """
        return SparseLinear(
            self.head_size,
            self.n_hidden,
            self.sparsity_mask,
            init_type=self.init_type,
            device=self.device,
        )

    def _prep_mask(self, mask: torch.Tensor) -> torch.Tensor:
        """
        Utility method. Preprocesses mask to match head size.

        !!! note "Performs three operations"

            1. Adds a padded matrix of 1s to end of mask in shape
                `(n_extras, n_extras)` where `n_extras=mask.shape[1]`
            2. Transposes mask from col matrix -> row matrix
            3. Gets the absolute values of the mask (swapping `-1 -> 1`)

        Parameters:
            mask (torch.Tensor): weight sparsity mask.

        Returns:
            mask (torch.Tensor): an updated mask.
        """
        n_extras = mask.shape[1]
        extra_nodes = torch.ones((n_extras, n_extras), device=self.device)
        mask = torch.concatenate([mask.detach(), extra_nodes])
        return torch.abs(mask.T).to(self.device)

    def _new_hidden(
        self, x: torch.Tensor, g_out: torch.Tensor, h_out: torch.Tensor
    ) -> torch.Tensor:
        """
        Helper method. Computes the new hidden state.

        Parameters:
            x (torch.Tensor): input values.
            g_out (torch.Tensor): g_head output.
            h_out (torch.Tensor): h_head output.

        Returns:
            hidden (torch.Tensor): a new hidden state
        """
        g_head = self.tanh(g_out)  # g(x, I, θ_g)
        h_head = self.tanh(h_out)  # h(x, I, θ_h)

        fh_g = self.f_head_to_g(x)
        fh_h = self.f_head_to_h(x)

        gate_out = self.sigmoid(fh_g + fh_h)  # [1 - σ(-[f(x, I, θf)], t)]
        f_head = 1.0 - gate_out  # σ(-f(x, I, θf), t)

        return g_head * f_head + gate_out * h_head

    def update_mask(self, mask: torch.Tensor) -> None:
        """
        Updates the sparsity mask with a new one.

        Parameters:
            mask (torch.Tensor): new mask
        """
        self.sparsity_mask = self._prep_mask(mask.to(self.device))

    def forward(
        self, x: torch.Tensor, hidden: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the cell.

        Parameters:
            x (torch.Tensor): input values.
            hidden (torch.Tensor): current hidden state.

        Returns:
            y_pred (torch.Tensor): the cell prediction.
            h_state (torch.Tensor): the hidden state.
        """
        x, hidden = x.to(self.device), hidden.to(self.device)
        x = torch.cat([x, hidden], dim=1)

        g_out = self.g_head(x)
        h_out = self.h_head(x)

        new_hidden = self._new_hidden(x, g_out, h_out)
        y_pred = self.proj(x) + new_hidden
        return y_pred, new_hidden



from velora.metrics.db import get_current_episode, get_db_engine
from velora.metrics.models import Episode, Experiment

__all__ = [
    "get_db_engine",
    "get_current_episode",
    "Experiment",
    "Episode",
]



from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Literal, Self, Tuple, Type

import gymnasium as gym
import torch
import torch.nn as nn
import torch.optim as optim

from velora.gym.wrap import add_core_env_wrappers
from velora.models.weight import WeightInitType
from velora.utils.core import set_seed
from velora.utils.torch import summary

if TYPE_CHECKING:
    from velora.buffer.base import BufferBase  # pragma: no cover
    from velora.models.nf.modules import (
        ActorModule,
        CriticModule,
        EntropyModule,
    )  # pragma: no cover

from velora.models.config import ModuleConfig, RLAgentConfig, TrainConfig
from velora.models.lnn.ncp import LiquidNCPNetwork, NCPNetwork

StateDictKeys = Literal["modules", "optimizers"]


class NCPModule(nn.Module):
    """
    A base class for NCP modules.

    Useful for Actor-Critic modules.
    """

    def __init__(
        self,
        in_features: int,
        n_neurons: int,
        out_features: int,
        *,
        init_type: str | WeightInitType = "kaiming_uniform",
        device: torch.device | None = None,
    ):
        """
        Parameters:
            in_features (int): the number of input nodes
            n_neurons (int): the number of hidden neurons
            out_features (int): the number of output nodes
            init_type (str, optional): the type of weight initialization
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__()

        self.in_features = in_features
        self.n_neurons = n_neurons
        self.out_features = out_features
        self.device = device

        self.ncp = NCPNetwork(
            in_features=in_features,
            n_neurons=n_neurons,
            out_features=out_features,
            init_type=init_type,
            device=device,
        ).to(device)

    def config(self) -> ModuleConfig:
        """
        Gets details about the module.

        Returns:
            config (ModuleConfig): a config model containing module details.
        """
        return ModuleConfig(
            active_params=self.ncp.active_params,
            total_params=self.ncp.total_params,
            architecture=summary(self),
        )


class LiquidNCPModule(nn.Module):
    """
    A base class for Liquid NCP modules.

    Useful for Actor-Critic modules.
    """

    def __init__(
        self,
        in_features: int,
        n_neurons: int,
        out_features: int,
        *,
        init_type: str | WeightInitType = "kaiming_uniform",
        device: torch.device | None = None,
    ):
        """
        Parameters:
            in_features (int): the number of input nodes
            n_neurons (int): the number of hidden neurons
            out_features (int): the number of output nodes
            init_type (str, optional): the type of weight initialization
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__()

        self.in_features = in_features
        self.n_neurons = n_neurons
        self.out_features = out_features
        self.device = device

        self.ncp = LiquidNCPNetwork(
            in_features=in_features,
            n_neurons=n_neurons,
            out_features=out_features,
            init_type=init_type,
            device=device,
        ).to(device)

    def config(self) -> ModuleConfig:
        """
        Gets details about the module.

        Returns:
            config (ModuleConfig): a config model containing module details.
        """
        return ModuleConfig(
            active_params=self.ncp.active_params,
            total_params=self.ncp.total_params,
            architecture=summary(self),
        )


class RLModuleAgent:
    """
    A base class for RL agents that use modules.

    Provides a blueprint describing the core methods that agents *must* have and
    includes useful utility methods.
    """

    def __init__(
        self,
        env: gym.Env,
        actor_neurons: int,
        critic_neurons: int,
        buffer_size: int,
        optim: Type[optim.Optimizer],
        device: torch.device | None,
        seed: int | None,
    ) -> None:
        """
        Parameters:
            env (gym.Env): Gymnasium environment to train on
            actor_neurons (int): number of decision nodes (inter and command nodes)
                for the actor
            critic_neurons (int): number of decision nodes (inter and command nodes)
                for the critic
            buffer_size (int): buffer capacity
            device (torch.device, optional): the device to perform computations on
            seed (int, optional): random number seed
        """
        self.env = env
        self.eval_env = add_core_env_wrappers(env, device)
        self.actor_neurons = actor_neurons
        self.critic_neurons = critic_neurons
        self.buffer_size = buffer_size
        self.optim = optim
        self.device = device
        self.seed = set_seed(seed)

        self.action_dim: int = (
            self.env.action_space.n.item()
            if isinstance(self.env.action_space, gym.spaces.Discrete)
            else self.env.action_space.shape[-1]
        )
        self.state_dim: int = self.env.observation_space.shape[-1]

        self.action_scale = None
        self.action_bias = None

        if isinstance(self.env.action_space, gym.spaces.Box):
            self.action_scale = (
                torch.tensor(
                    self.env.action_space.high - self.env.action_space.low,
                    device=device,
                )
                / 2.0
            )
            self.action_bias = (
                torch.tensor(
                    self.env.action_space.high + self.env.action_space.low,
                    device=device,
                )
                / 2.0
            )

        self.config: RLAgentConfig | None = None
        self.buffer: "BufferBase" | None = None

        self.actor: "ActorModule" | None = None
        self.critic: "CriticModule" | None = None

        self.entropy: "EntropyModule" | None = None

        self.active_params = 0
        self.total_params = 0

        self.metadata: Dict[str, Any] = {}

    @abstractmethod
    def train(
        self,
        n_episodes: int,
        max_steps: int,
        window_size: int,
        *args,
        **kwargs,
    ) -> Any:
        pass  # pragma: no cover

    @abstractmethod
    def predict(
        self,
        state: torch.Tensor,
        hidden: torch.Tensor,
        train_mode: bool = False,
        *args,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        pass  # pragma: no cover

    @abstractmethod
    def save(
        self,
        dirpath: str | Path,
        *,
        buffer: bool = False,
        config: bool = False,
    ) -> None:
        """
        Saves the current model state into `safetensors` and `json` files.

        !!! warning

            `model_config.json` is stored in the `dirpath.parent`.

        Includes:

        - `model_config.json` - contains the core details of the agent (optional)
        - `metadata.json` - contains the model, optimizer and buffer (optional) metadata
        - `model_state.safetensors` - contains the model weights and biases
        - `optim_state.safetensors` - contains the optimizer states (actor and critic)
        - `buffer_state.safetensors` - contains the buffer state (only if `buffer=True`)

        Parameters:
            dirpath (str | Path): the location to store the model state. Should only
                consist of `folder` names. E.g., `<folder>/<folder>`
            buffer (bool, optional): a flag for storing the buffer state
            config (bool, optional): a flag for storing the model's config
        """
        pass  # pragma: no cover

    @classmethod
    @abstractmethod
    def load(cls, dirpath: str | Path, *, buffer: bool = False) -> Self:
        """
        Creates a new agent instance by loading a saved one from the `dirpath`.
        Also, loads the original training buffer if `buffer=True`.

        These files must exist in the `dirpath`:

        - `metadata.json` - contains the model, optimizer and buffer (optional) metadata
        - `model_state.safetensors` - contains the model weights and biases
        - `optim_state.safetensors` - contains the optimizer states (actor and critic)
        - `buffer_state.safetensors` - contains the buffer state (only if `buffer=True`)

        Parameters:
            dirpath (str | Path): the location to store the model state. Should only
                consist of `folder` names. E.g., `<folder>/<folder>`
            buffer (bool, optional): a flag for storing the buffer state

        Returns:
            agent (Self): a new agent instance with the saved state
        """
        pass  # pragma: no cover

    def state_dict(self) -> Dict[StateDictKeys, Dict[str, Any]]:
        """
        Retrieves the agent's module state dictionaries and splits them into
        categories.

        Returns:
            state_dict (Dict[Literal["modules", "optimizers"], Dict[str, Any]]): the agent's module state dicts categorized.
        """
        final_dict: Dict[StateDictKeys, Dict[str, Any]] = {
            "modules": {},
            "optimizers": {},
        }

        for module in [self.actor, self.critic, self.entropy]:
            if module is not None:
                state_dict: Dict[str, Any] = module.state_dict()

                for key, val in state_dict.items():
                    if "optim" in key:
                        final_dict["optimizers"][key] = val
                    else:
                        final_dict["modules"][key] = val

        return final_dict

    def set_metadata(self, values: Dict[str, Any], seed: int) -> Dict[str, Any]:
        """
        Creates the agents metadata based on a given set of local variables.

        Parameters:
            values (Dict[str, Any]): local variables
            seed (int): randomly generated seed

        Returns:
            metadata (Dict[str, Any]): an updated dictionary of agent metadata.
        """
        metadata = {
            k: v for k, v in values.items() if k not in ["self", "__class__", "env"]
        }
        metadata["device"] = str(self.device) if self.device is not None else "cpu"
        metadata["optim"] = f"torch.optim.{self.optim.__name__}"
        metadata["seed"] = seed

        return metadata

    def _set_train_params(self, params: Dict[str, Any]) -> TrainConfig:
        """
        Helper method. Sets the `train_params` given a dictionary of training parameters.

        Parameters:
            params (Dict[str, Any]): a dictionary of training parameters

        Returns:
            config (TrainConfig): a training config model
        """
        params = dict(
            callbacks=(
                dict(cb.config() for cb in params["callbacks"])
                if params["callbacks"]
                else None
            ),
            **{k: v for k, v in params.items() if k not in ["self", "callbacks"]},
        )
        return TrainConfig(**params)



from enum import Enum
from typing import Literal

import torch
import torch.nn as nn


class LeCunTanh(nn.Module):
    """
    Implements LeCun's Tanh activation function.
    $$
    f(x) = 1.7159 \\tanh (\\frac{2}{3} x)
    $$
    Constants are applied to keep the variance of the output close to `1`.
    """

    def __init__(self) -> None:
        super().__init__()

        self.tanh = nn.Tanh()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return 1.7159 * self.tanh(0.666 * x)


ActivationTypeLiteral = Literal[
    "relu",
    "tanh",
    "elu",
    "leaky_relu",
    "prelu",
    "selu",
    "silu",
    "softsign",
    "sigmoid",
    "hardsigmoid",
    "lecun_tanh",
]


class ActivationEnum(Enum):
    """
    An Enum for PyTorch activation functions.

    Useful for getting activation functions dynamically using a `string` name.
    Refer to the `get()` method for more details.
    """

    RELU = nn.ReLU()
    TANH = nn.Tanh()
    ELU = nn.ELU()
    LEAKY_RELU = nn.LeakyReLU()
    PRELU = nn.PReLU()
    SELU = nn.SELU()
    SILU = nn.GELU()
    SOFTSIGN = nn.Softsign()
    SIGMOID = nn.Sigmoid()
    HARDSIGMOID = nn.Hardsigmoid()
    LECUN_TANH = LeCunTanh()

    @classmethod
    def get(cls, name: ActivationTypeLiteral) -> nn.Module:
        """
        Get the `torch.nn` activation function.

        Parameters:
            name (Literal["relu", "tanh", "elu", "leaky_relu", "prelu", "selu", "silu", "softsign", "sigmoid", "hardsigmoid", "lecun_tanh"]):
                the name of the activation function.

        Returns:
            activation (nn.Module): the PyTorch activation module.
        """
        try:
            return cls[name.upper()].value
        except KeyError:
            raise ValueError(f"Unsupported activation function: {name}")


__all__ = [
    "LeCunTanh",
    "ActivationTypeLiteral",
    "ActivationEnum",
]



from dataclasses import dataclass

import torch


@dataclass
class BatchExperience:
    """
    Storage container for a batch agent experiences.

    Attributes:
        states (torch.Tensor): a batch of environment observations
        actions (torch.Tensor): a batch of agent actions taken in the states
        rewards (torch.Tensor): a batch of rewards obtained for taking the actions
        next_states (torch.Tensor): a batch of newly generated environment
            observations following the actions taken
        dones (torch.Tensor): a batch of environment completion statuses
        hiddens (torch.Tensor): a batch of prediction network hidden states
            (e.g., Actor)
    """

    states: torch.Tensor
    actions: torch.Tensor
    rewards: torch.Tensor
    next_states: torch.Tensor
    dones: torch.Tensor
    hiddens: torch.Tensor



import os
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, get_args

try:
    from typing import override
except ImportError:  # pragma: no cover
    from typing_extensions import override  # pragma: no cover

import gymnasium as gym

if TYPE_CHECKING:
    from velora.state import TrainState  # pragma: no cover
    from velora.models.base import RLModuleAgent  # pragma: no cover

from velora.metrics.db import get_current_episode
from velora.state import AnalyticsState, RecordMethodLiteral, RecordState
from velora.utils.format import number_to_short


class TrainCallback:
    """
    Abstract base class for all training callbacks.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def __call__(self, state: "TrainState") -> "TrainState":
        """
        The callback function that gets called during training.

        Parameters:
            state (TrainState): the current training state

        Returns:
            state (TrainState): the current training state.
        """
        pass  # pragma: no cover

    @abstractmethod
    def config(self) -> Tuple[str, Dict[str, Any]]:
        """
        Retrieves callback details in the form: `(name, values)`.

        Returns:
            name (str): callback name.
            values (Dict[str, Any]): a dictionary containing callback settings.
        """
        pass  # pragma: no cover

    @abstractmethod
    def info(self) -> str:
        """
        Provides details with basic information about the callback initialization.

        Returns:
            details (str): a string of information.
        """
        pass  # pragma: no cover


class EarlyStopping(TrainCallback):
    """
    A callback that applies early stopping to the training process.
    """

    @override
    def __init__(self, target: float, *, patience: int = 3) -> None:
        """
        Parameters:
            target (float): episode reward target to achieve
            patience (int, optional): number of times the threshold needs
                to be met to terminate training
        """
        self.target = target
        self.patience = patience

        self.count = 0

    def __call__(self, state: "TrainState") -> "TrainState":
        if state.stop_training:
            return state

        if state.status == "episode":
            if state.ep_reward >= self.target:
                self.count += 1

                if self.count >= self.patience:
                    state.stop_training = True
            else:
                self.count = 0

        return state

    def config(self) -> Tuple[str, Dict[str, Any]]:
        return self.__class__.__name__, {
            "target": self.target,
            "patience": self.patience,
        }

    def info(self) -> str:
        return f"'{self.__class__.__name__}' enabled with 'target={self.target}' and 'patience={self.patience}'."


class SaveCheckpoints(TrainCallback):
    """
    A callback that applies model state saving checkpoints to the training process.
    """

    @override
    def __init__(
        self,
        dirname: str,
        *,
        frequency: int = 100,
        buffer: bool = False,
    ) -> None:
        """
        Parameters:
            dirname (str): the model directory name to save checkpoints.
                Automatically created inside `checkpoints` directory as
                `checkpoints/<dirname>/saves`.

                Compliments `TrainCallback.RecordVideos` callback.

            frequency (int, optional): save frequency (in episodes or steps)
            buffer (bool, optional): whether to save the buffer state
        """
        self.dirname = dirname
        self.frequency = frequency
        self.buffer = buffer

        self.filepath = Path("checkpoints", self.dirname, "saves")

        if self.filepath.exists():
            raise FileExistsError(
                f"Items already exist in the '{self.filepath.parent}' directory! Either change the 'dirname' or delete the folders contents."
            )

    def __call__(self, state: "TrainState") -> "TrainState":
        if state.status == "start":
            state.saving_enabled = True
            state.checkpoint_dir = self.filepath

        # Only perform checkpoint operations on episode and complete events
        if state.status != "episode" and state.status != "complete":
            return state

        # Create directory if it doesn't exist on first call
        self.filepath.mkdir(parents=True, exist_ok=True)

        ep_idx = state.current_ep

        should_save = False
        filename = f"{state.env.spec.name}_"

        # Save checkpoint at specified frequency
        if state.status == "episode" and ep_idx != state.total_episodes:
            if ep_idx % self.frequency == 0:
                should_save = True
                filename += f"{number_to_short(ep_idx)}"

        # Perform final checkpoint save
        elif state.status == "complete":
            should_save = True
            filename += "final"

        if should_save:
            self.save_checkpoint(state.agent, filename)

        return state

    def save_checkpoint(self, agent: "RLModuleAgent", dirname: str) -> None:
        """
        Saves a checkpoint at a given episode with the given suffix.

        Parameters:
            agent (RLModuleAgent): the agent being trained
            dirname (str): the checkpoint directory name
        """
        checkpoint_path = Path(self.filepath, dirname)
        agent.save(checkpoint_path, buffer=self.buffer, config=True)

    def config(self) -> Tuple[str, Dict[str, Any]]:
        return self.__class__.__name__, {
            "dirname": self.dirname,
            "frequency": self.frequency,
            "buffer": self.buffer,
        }

    def info(self) -> str:
        return (
            f"'{self.__class__.__name__}' enabled with 'frequency={self.frequency}' and 'buffer={self.buffer}'.\n"
            f"    Files will be saved in the 'checkpoints/{self.dirname}/saves' directory."
        )


class RecordVideos(TrainCallback):
    """
    A callback to enable intermittent environment video recording to visualize
    the agent training progress.

    Requires environment with `render_mode="rgb_array"`.
    """

    @override
    def __init__(
        self,
        dirname: str,
        *,
        method: RecordMethodLiteral = "episode",
        frequency: int = 100,
        force: bool = False,
    ) -> None:
        """
        Parameters:
            dirname (str): the model directory name to store the videos.
                Automatically created in `checkpoints` directory as
                `checkpoints/<dirname>/videos`.

                Compliments `TrainCallback.SaveCheckpoints` callback.

            method (Literal["episode", "step"], optional): the recording method.
                When `episode` records episodically. When `step` records during
                training steps.
            frequency (int, optional): the `episode` or `step` record frequency
            force (bool, optional): enables file overwriting, ignoring existing
                video files. Useful for continuing model training
        """
        if method not in get_args(RecordMethodLiteral):
            raise ValueError(
                f"'{method=}' is not supported. Choices: '{get_args(RecordMethodLiteral)}'"
            )

        self.dirname = dirname
        self.method = method
        self.frequency = frequency
        self.force = force

        self.dirpath = Path("checkpoints", dirname, "videos")

        if not force and self.dirpath.exists():
            raise FileExistsError(
                f"Files already exist in the '{self.dirpath.parent}' directory! Either change the 'dirname', delete/move the folder and its contents, or use 'force=True' to allow overwriting."
            )

        def trigger(t: int) -> bool:
            # Skip first item
            if t == 0:
                return False

            return t % frequency == 0

        self.details = RecordState(
            dirpath=self.dirpath,
            method=method,
            episode_trigger=trigger if method == "episode" else None,
            step_trigger=trigger if method == "step" else None,
        )

    def __call__(self, state: "TrainState") -> "TrainState":
        # 'start': Set the recording state
        if state.status == "start":
            state.record_state = self.details

            state.env = gym.wrappers.RecordVideo(
                state.env,
                name_prefix=state.env.spec.name,
                **state.record_state.to_wrapper(),
            )

        # Ignore other events
        return state

    def config(self) -> Tuple[str, Dict[str, Any]]:
        return self.__class__.__name__, {
            "dirname": self.dirname,
            "method": self.method,
            "frequency": self.frequency,
            "force": self.force,
        }

    def info(self) -> str:
        return (
            f"'{self.__class__.__name__}' enabled with 'method={str(self.method)}' and 'frequency={self.frequency}'.\n"
            f"    Files will be saved in the 'checkpoints/{self.dirname}/videos' directory."
        )


class CometAnalytics(TrainCallback):
    """
    A callback that enables [`comet-ml`](https://www.comet.com/site/) cloud-based
    analytics tracking.

    Requires Comet ML API key set using the `COMET_API_KEY` environment variable.

    Features:

    - Upload agent configuration objects
    - Tracks episodic training metrics
    - Uploads video recordings (if `RecordVideos` callback applied)
    """

    @override
    def __init__(
        self,
        project_name: str,
        experiment_name: str | None = None,
        *,
        tags: List[str] | None = None,
        experiment_key: str | None = None,
    ) -> None:
        """
        Parameters:
            project_name (str): the name of the Comet ML project to add this
                experiment to
            experiment_name (str, optional): the name of this experiment run.
                If `None`, automatically creates the name using the format
                `<agent_classname>_<env_name>_<n_episodes>ep`. Ignored when
                using `experiment_key`
            tags (List[str], optional): a list of tags associated with the
                experiment. If `None` adds the `agent_classname` and `env_name`
                by default. Ignored when using `experiment_key`
            experiment_key (str, optional): an existing Comet ML experiment key.
                Used for continuing an experiment
        """
        try:
            from comet_ml import Experiment
        except ImportError:
            raise ImportError(
                "Failed to load the 'comet_ml' package. Have you installed it using 'pip install velora[comet]'?"
            )

        api_key = os.getenv("COMET_API_KEY", None)
        if api_key is None or api_key == "":
            raise ValueError(
                "Missing 'api_key'! Store it as a 'COMET_API_KEY' environment variable."
            )

        self.experiment: Experiment | None = None
        self.experiment_key = experiment_key

        self.state = AnalyticsState(
            project_name=project_name,
            experiment_name=experiment_name,
            tags=tags,
        )

        self.project_name = project_name
        self.experiment_name = experiment_name if experiment_name else "auto"
        self.tags = tags if tags else "auto"

    def __call__(self, state: "TrainState") -> "TrainState":
        # Setup experiment
        if state.status == "start":
            # Update comet training state
            state.analytics_state = self.state
            state.analytics_update()

            self.init_experiment(state)

            # Log config
            self.experiment.log_parameters(state.agent.config.model_dump())

        # Send episodic metrics
        if state.status == "logging":
            if state.logging_type == "episode":
                self.log_episode_data(state)

        # Finalize training
        if state.status == "complete":
            # Log video recordings
            if state.record_state is not None:
                for video in state.record_state.dirpath.iterdir():
                    self.experiment.log_video(str(video), format="mp4")

            self.experiment.end()

        return state

    def log_episode_data(self, state: "TrainState") -> None:
        """
        Logs episodic data to the experiment.

        Parameters:
            state (TrainState): the current training state
        """
        results = get_current_episode(
            state.session,
            state.experiment_id,
            state.current_ep,
        )

        for item in results:
            reward_low = item.reward_moving_avg - item.reward_moving_std
            reward_high = item.reward_moving_avg + item.reward_moving_std

            self.experiment.log_metrics(
                {
                    "reward/moving_lower": reward_low,
                    "reward/moving_avg": item.reward_moving_avg,
                    "reward/moving_upper": reward_high,
                    "episode/return": item.reward,
                    "episode/length": item.length,
                    "losses/actor": item.actor_loss,
                    "losses/critic": item.critic_loss,
                    "losses/entropy": item.entropy_loss,
                },
                epoch=state.current_ep,
            )

    def init_experiment(self, state: "TrainState") -> None:
        """Setups up a comet experiment and stores it locally.

        Parameters:
            state (TrainState): the current training state
        """
        from comet_ml import ExistingExperiment, Experiment

        if self.experiment_key is not None:
            # Continue existing experiment
            self.experiment = ExistingExperiment(
                api_key=os.getenv("COMET_API_KEY", None),
                project_name=state.analytics_state.project_name,
                experiment_key=self.experiment_key,
                disabled=bool(os.getenv("VELORA_TEST_MODE", "").lower())
                in ("true", "1"),
                log_env_cpu=True,
                log_env_gpu=True,
                log_env_details=True,
            )
        else:
            # Start new experiment
            self.experiment = Experiment(
                api_key=os.getenv("COMET_API_KEY", None),
                project_name=state.analytics_state.project_name,
                auto_param_logging=False,
                auto_metric_logging=False,
                auto_output_logging=False,
                log_graph=False,
                display_summary_level=0,
                disabled=bool(os.getenv("VELORA_TEST_MODE", "").lower())
                in ("true", "1"),
            )

            self.experiment.set_name(state.analytics_state.experiment_name)
            self.experiment.add_tags(state.analytics_state.tags)

    def config(self) -> Tuple[str, Dict[str, Any]]:
        return self.__class__.__name__, {
            "project_name": self.project_name,
            "experiment_name": self.experiment_name,
            "tags": ",".join(self.tags) if isinstance(self.tags, list) else self.tags,
            "experiment_key": self.experiment_key,
        }

    def info(self) -> str:
        return f"'{self.__class__.__name__}' enabled with 'project_name={self.project_name}', 'experiment_name={self.experiment_name}' and 'tags={self.tags}'."
